import collections.abc as cabc
import copy
import itertools
import typing as typ
import warnings

import numpy as np
import numpy.random as npr
import pytest
import scipy.linalg as splinalg
import scipy.sparse.linalg as spsl

import pyxu.abc as pxa
import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.info.warning as pxw
import pyxu.math as pxm
import pyxu.runtime as pxrt
import pyxu.util as pxu
import pyxu_tests.conftest as ct
from pyxu.abc.operator import _core_operators


def get_test_class(cls: pxt.OpC) -> "MapT":
    # Find the correct MapT subclass designed to test `cls`.
    is_test = lambda _: all(
        [
            hasattr(_, "base"),
            hasattr(_, "interface"),
        ]
    )
    candidates = {_ for _ in globals().values() if is_test(_)}
    for clsT in candidates:
        if clsT.base == cls:
            return clsT
    else:
        raise ValueError(f"No known test type for {cls}.")


# Naming conventions
# ------------------
#
# * data_<property>()
#       Return expected object of `op.<property>`.
#
# * test_<property>(op, ...):
#       Verify property values.
#
# * data_<method>(op, ...)
#       Return mappings of the form dict(in_=dict(), out=Any), where:
#         * in_ are kwargs to `op.<method>()`;
#         * out denotes the output of `op.method(**data[in_])`.
#
# * data_math_<method>()
#       Special test data for mathematical identities.
#
# * test_[value,backend,
#         prec,precCM,
#         transparent,math,
#         interface
#        ]_<method>(op, ...)
#       Verify that <method>, returns
#       * value: right output values
#       * backend: right output type
#       * prec: input/output have same precision
#       * precCM: output respects context-manager choice
#       * transparent: referential-transparency, i.e. no side-effects
#       * math: mathematical identities hold
#       * interface: objects have the right interface
#
DataLike = cabc.Mapping[str, typ.Any]


class MapT(ct.DisableTestMixin):
    # Class Properties --------------------------------------------------------
    base = pxa.Map
    interface: cabc.Set[str] = frozenset(
        {
            "shape",
            "dim",
            "codim",
            "asop",
            "argshift",
            "argscale",
            "apply",
            "__call__",
            "lipschitz",
            "estimate_lipschitz",
            "expr",
            "squeeze",
        }
    )

    # Internal helpers --------------------------------------------------------
    @staticmethod
    def _skip_if_unsupported(ndi: pxd.NDArrayInfo):
        if ndi.module() is None:
            pytest.skip(f"{ndi} unsupported on this machine.")

    @staticmethod
    def _random_array(
        shape: pxt.NDArrayShape,
        seed: int = 0,
        xp: pxt.ArrayModule = pxd.NDArrayInfo.NUMPY.module(),
        width: pxrt.Width = pxrt.Width.DOUBLE,
    ) -> pxt.NDArray:
        rng = npr.default_rng(seed)
        x = rng.normal(size=shape)
        return xp.array(x, dtype=width.value)

    @staticmethod
    def _sanitize(x, default):
        if x is not None:
            return x
        else:
            return default

    @staticmethod
    def _check_has_interface(op: pxa.Map, klass: "MapT"):
        # Verify `op` has the public interface of `klass`.
        assert klass.interface <= frozenset(dir(op))

    @classmethod
    def _metric(
        cls,
        a: pxt.NDArray,
        b: pxt.NDArray,
        as_dtype: pxt.DType,
    ) -> bool:
        # Function used to assess if computed values returned by arithmetic methods are correct.
        #
        # Users may override this function to introduce an alternative metric when justified.
        # The default metric is point-wise match.
        #
        # Parameters
        # ----------
        # a, b: pxt.NDArray
        #    (..., N) arrays
        # as_dtype: pxt.DType
        #    dtype used to compare the values. (Not always relevant depending on the metric.)
        #
        # Returns
        # -------
        # match: bool
        #    True if all (...) arrays match.
        return ct.allclose(a, b, as_dtype)

    @classmethod
    def _check_value1D(
        cls,
        func,
        data: DataLike,
        dtype: pxt.DType = None,
    ):
        in_ = data["in_"]
        with pxrt.EnforcePrecision(False):
            out = func(**in_)
        out_gt = data["out"]

        dtype = MapT._sanitize(dtype, in_["arr"].dtype)
        assert out.ndim == in_["arr"].ndim
        assert cls._metric(out, out_gt, as_dtype=dtype)

    @classmethod
    def _check_valueND(
        cls,
        func,
        data: DataLike,
        dtype: pxt.DType = None,
    ):
        sh_extra = (2, 1, 3)  # prepend input/output shape by this amount.

        in_ = data["in_"]
        arr = in_["arr"]
        xp = pxu.get_array_module(arr)
        arr = xp.broadcast_to(arr, (*sh_extra, *arr.shape))
        in_.update(arr=arr)
        with pxrt.EnforcePrecision(False):
            out = func(**in_)
        out_gt = np.broadcast_to(data["out"], (*sh_extra, *data["out"].shape))

        dtype = MapT._sanitize(dtype, in_["arr"].dtype)
        assert out.ndim == in_["arr"].ndim
        assert cls._metric(out, out_gt, as_dtype=dtype)

    @staticmethod
    def _check_backend(func, data: DataLike):
        in_ = data["in_"]
        with pxrt.EnforcePrecision(False):
            out = func(**in_)

        assert type(out) == type(in_["arr"])  # noqa: E721

    @staticmethod
    def _check_prec(func, data: DataLike):
        in_ = data["in_"]
        with pxrt.EnforcePrecision(False):
            out = func(**in_)
            assert out.dtype == in_["arr"].dtype

    @staticmethod
    def _check_precCM(
        func,
        data: DataLike,
        widths: cabc.Collection[pxrt.Width] = pxrt.Width,
    ):
        stats = dict()
        for w in widths:
            with pxrt.Precision(w):
                out = func(**data["in_"])
            stats[w] = out.dtype == w.value
        assert all(stats.values())

    @staticmethod
    def _check_precCM_func(func, width: pxrt.Width):
        with pxrt.Precision(width):
            out = func()
        assert out.dtype == width.value

    @classmethod
    def _check_no_side_effect(cls, func, data: DataLike):
        # idea:
        # * eval func() once [out_1]
        # * in-place update out_1, ex: scale by constant factor [scale]
        # * re-eval func() [out_2]
        # * assert out_1 == out_2, i.e. input was not modified
        data = copy.deepcopy(data)
        in_, out_gt = data["in_"], data["out"]
        scale = 10
        out_gt *= scale

        with pxrt.EnforcePrecision(False):
            out_1 = func(**in_)
            if pxu.copy_if_unsafe(out_1) is not out_1:
                # out_1 is read_only -> safe
                return
            else:
                # out_1 is writeable -> test correctness
                out_1 *= scale
                out_2 = func(**in_)
                out_2 *= scale

            try:
                # The large scale introduced to assess transparency may give rise to
                # operator-dependant round-off errors. We therefore assess transparency at
                # FP32-precision to avoid false negatives.
                assert cls._metric(out_1, out_gt, as_dtype=pxrt.Width.SINGLE.value)
                assert cls._metric(out_2, out_gt, as_dtype=pxrt.Width.SINGLE.value)
            except AssertionError:
                # Function is non-transparent, but which backend caused it?
                N = pxd.NDArrayInfo
                ndi = N.from_obj(out_1)
                if ndi == N.CUPY:
                    # warn about CuPy-only non-transparency.
                    msg = "\n".join(
                        [
                            f"{func} is not transparent when applied to CuPy inputs.",
                            f"If the same test fails for non-CuPy inputs, then {func}'s implementation is at fault -> user fix required.",
                            "If the same test passes for non-CuPy inputs, then this warning can be safely ignored.",
                        ]
                    )
                    warnings.warn(msg, pxw.NonTransparentWarning)
                else:
                    raise

    # Fixtures ----------------------------------------------------------------
    @pytest.fixture
    def spec(self) -> tuple[pxt.OpT, pxd.NDArrayInfo, pxrt.Width]:
        # override in subclass to return:
        # * the operator to test;
        # * the backend of accepted input arrays;
        # * the precision of accepted input arrays.
        #
        # The triplet (op, backend, precision) must be provided since some operators may not be
        # backend/precision-agnostic.
        raise NotImplementedError

    @pytest.fixture
    def op(self, spec) -> pxt.OpT:
        return spec[0]

    @pytest.fixture
    def ndi(self, spec) -> pxd.NDArrayInfo:
        ndi_ = spec[1]
        self._skip_if_unsupported(ndi_)
        return ndi_

    @pytest.fixture
    def xp(self, ndi) -> pxt.ArrayModule:
        return ndi.module()

    @pytest.fixture
    def _gpu(self, ndi) -> bool:
        # Boolean flag needed by some methods
        return ndi == pxd.NDArrayInfo.CUPY

    @pytest.fixture
    def width(self, spec) -> pxrt.Width:
        return spec[2]

    @pytest.fixture(params=_core_operators())
    def _klass(self, request) -> pxt.OpC:
        # Returns some operator in the Operator hierarchy.
        # Do not override in subclass: for internal use only to test `op.asop()`.
        return request.param

    @pytest.fixture
    def data_shape(self) -> pxt.OpShape:
        # override in subclass with the shape of op.
        # Don't return `op.shape`: hard-code what you are expecting.
        raise NotImplementedError

    @pytest.fixture
    def data_apply(self) -> DataLike:
        # override in subclass with 1D input/outputs of op.apply().
        # Arrays should be NumPy-only. (Internal machinery will transform to different
        # backend/precisions as needed.)
        raise NotImplementedError

    @pytest.fixture
    def data_math_lipschitz(self) -> cabc.Collection[np.ndarray]:
        # override in subclass with at least 2 evaluation points for op.apply().
        # Used to verify if op.apply() satisfies the Lipschitz condition.
        # Arrays should be NumPy-only. (Internal machinery will transform to different
        # backend/precisions as needed.)
        raise NotImplementedError

    @pytest.fixture
    def _data_estimate_lipschitz(self) -> DataLike:
        # Generate Cartesian product of inputs.
        # For internal use only to test `op.test_math_lipschitz()`.
        data = dict(in_=dict())
        return data

    @pytest.fixture
    def _data_apply(self, data_apply, xp, width) -> DataLike:
        # Generate Cartesian product of inputs.
        # Do not override in subclass: for internal use only to test `op.apply()`.
        # Outputs are left unchanged: different tests should transform them as required.
        in_ = copy.deepcopy(data_apply["in_"])
        in_.update(arr=xp.array(in_["arr"], dtype=width.value))
        data = dict(
            in_=in_,
            out=data_apply["out"],
        )
        return data

    # Tests -------------------------------------------------------------------
    def test_interface(self, op):
        self._skip_if_disabled()
        self._check_has_interface(op, self.__class__)

    def test_shape(self, op, data_shape):
        self._skip_if_disabled()
        assert op.shape == data_shape

    def test_dim(self, op, data_shape):
        self._skip_if_disabled()
        assert op.dim == data_shape[1]

    def test_codim(self, op, data_shape):
        self._skip_if_disabled()
        assert op.codim == data_shape[0]

    def test_value1D_apply(self, op, _data_apply):
        self._skip_if_disabled()
        self._check_value1D(op.apply, _data_apply)

    def test_valueND_apply(self, op, _data_apply):
        self._skip_if_disabled()
        self._check_valueND(op.apply, _data_apply)

    def test_backend_apply(self, op, _data_apply):
        self._skip_if_disabled()
        self._check_backend(op.apply, _data_apply)

    def test_prec_apply(self, op, _data_apply):
        self._skip_if_disabled()
        self._check_prec(op.apply, _data_apply)

    def test_precCM_apply(self, op, _data_apply):
        self._skip_if_disabled()
        self._check_precCM(op.apply, _data_apply)

    def test_transparent_apply(self, op, _data_apply):
        self._skip_if_disabled()
        self._check_no_side_effect(op.apply, _data_apply)

    def test_value1D_call(self, op, _data_apply):
        self._skip_if_disabled()
        self._check_value1D(op.__call__, _data_apply)

    def test_valueND_call(self, op, _data_apply):
        self._skip_if_disabled()
        self._check_valueND(op.__call__, _data_apply)

    def test_backend_call(self, op, _data_apply):
        self._skip_if_disabled()
        self._check_backend(op.__call__, _data_apply)

    def test_prec_call(self, op, _data_apply):
        self._skip_if_disabled()
        self._check_prec(op.__call__, _data_apply)

    def test_precCM_call(self, op, _data_apply):
        self._skip_if_disabled()
        self._check_precCM(op.__call__, _data_apply)

    def test_transparent_call(self, op, _data_apply):
        self._skip_if_disabled()
        self._check_no_side_effect(op.__call__, _data_apply)

    def test_interface_lipschitz(self, op):
        # .lipschitz() always returns a float.
        self._skip_if_disabled()
        L = op.lipschitz
        assert isinstance(L, pxt.Real)

    def test_interface2_lipschitz(self, op):
        # .lipschitz() can be manually set.
        # [and negative values are illegal.]
        self._skip_if_disabled()
        L_orig = op.lipschitz

        L_new = self._random_array((1,)).item()
        op.lipschitz = L_new
        assert isinstance(op.lipschitz, pxt.Real)
        with pxrt.EnforcePrecision(False):
            assert np.isclose(op.lipschitz, L_new)

        with pytest.raises(Exception):
            op.lipschitz = -1

        # reset state for next tests
        op.lipschitz = L_orig

    def test_precCM_lipschitz(self, op, width):
        self._skip_if_disabled()
        func = lambda: op.lipschitz
        self._check_precCM_func(func, width)

    # We disable RuntimeWarnings which may arise due to NaNs. (See comment below.)
    @pytest.mark.filterwarnings("ignore::RuntimeWarning")
    def test_math_lipschitz(
        self,
        op,
        xp,
        width,
        data_math_lipschitz,
        _data_estimate_lipschitz,
    ):
        # \norm{f(x) - f(y)}{2} \le L * \norm{x - y}{2}
        # L provided by `estimate_lipschitz()`.
        self._skip_if_disabled()
        with pxrt.EnforcePrecision(False):
            L = op.estimate_lipschitz(**_data_estimate_lipschitz["in_"])

            data = xp.array(data_math_lipschitz, dtype=width.value)
            data = list(itertools.combinations(data, 2))
            x = xp.stack([_[0] for _ in data], axis=0)
            y = xp.stack([_[1] for _ in data], axis=0)

            lhs = pxm.norm(op.apply(x) - op.apply(y), axis=-1)
            # .apply() may return INFs, in which case `INF-INF=NaN` may arise above.
            # less_equal() is not able to handle NaNs, so the former are overwritten by a sensible
            # value in this context, i.e. 0.
            lhs[xp.isnan(lhs)] = 0

            rhs = L * pxm.norm(x - y, axis=-1)
            success = ct.less_equal(lhs, rhs, as_dtype=width.value)
            assert all(success)

    def test_interface_asop(self, op, _klass):
        # * .asop() is no-op if same klass or parent
        # * .asop() has correct interface otherwise.
        # * Expect an exception if cast is illegal. (shape-issues)
        self._skip_if_disabled()
        P = pxa.Property
        if _klass in op.__class__.__mro__:
            op2 = op.asop(_klass)
            assert op2 is op
        elif (P.FUNCTIONAL in _klass.properties()) and (op.codim > 1):
            # Casting to functionals when codim != 1.
            with pytest.raises(Exception):
                op2 = op.asop(_klass)
        elif (P.LINEAR_SQUARE in _klass.properties()) and (op.codim != op.dim):
            # Casting non-square operators to square linops.
            with pytest.raises(Exception):
                op2 = op.asop(_klass)
        else:
            op2 = op.asop(_klass)
            klassT = get_test_class(_klass)
            self._check_has_interface(op2, klassT)

    def test_value_asop(self, op, _klass):
        # Ensure encapsulated arithmetic fields are forwarded.
        # We only test arithmetic fields known to belong to any Map subclass:
        # * estimate_lipschitz (method)
        # * apply (method)
        # * __call__ (special method)
        self._skip_if_disabled()
        try:
            op2 = op.asop(_klass)
        except Exception:
            # nothing to test since `op.asop(_klass)` is illegal.
            return
        else:
            assert op2.estimate_lipschitz == op.estimate_lipschitz
            assert op2.apply == op.apply
            assert op2.__call__ == op.__call__


class FuncT(MapT):
    # Class Properties --------------------------------------------------------
    base = pxa.Func
    interface = frozenset(MapT.interface | {"asloss"})

    # Tests -------------------------------------------------------------------
    def test_codim(self, op):
        self._skip_if_disabled()
        assert op.codim == 1

    def test_interface_asloss(self, op, xp, width):
        # op.asloss() sub-classes Func if data provided, transparent otherwise.
        # Disable this test if asloss() not defined for a functional.
        self._skip_if_disabled()

        N_dim = self._sanitize(op.dim, default=3)
        data = self._random_array((N_dim,), xp=xp, width=width)
        try:
            assert op.asloss() is op
            self._check_has_interface(op.asloss(data), self.__class__)
        except NotImplementedError:
            msg = f"{op.asloss} is undefined, but `test_interface_asloss()` was not disabled."
            assert False, msg


class DiffMapT(MapT):
    # Class Properties --------------------------------------------------------
    base = pxa.DiffMap
    interface = frozenset(
        MapT.interface
        | {
            "jacobian",
            "diff_lipschitz",
            "estimate_diff_lipschitz",
        }
    )

    # Fixtures ----------------------------------------------------------------
    @pytest.fixture
    def data_math_diff_lipschitz(self) -> cabc.Collection[np.ndarray]:
        # override in subclass with at least 2 evaluation points for op.apply().
        # Used to verify if op.jacobian() satisfies the diff_Lipschitz condition.
        # Arrays should be NumPy-only. (Internal machinery will transform to different
        # backend/precisions as needed.)
        raise NotImplementedError

    @pytest.fixture
    def _data_estimate_diff_lipschitz(self) -> DataLike:
        # Generate Cartesian product of inputs.
        # For internal use only to test `op.test_math_diff_lipschitz()`.
        data = dict(in_=dict())
        return data

    # Tests -------------------------------------------------------------------
    def test_interface_jacobian(self, op, _data_apply):
        self._skip_if_disabled()
        arr = _data_apply["in_"]["arr"]
        J = op.jacobian(arr)
        self._check_has_interface(J, LinOpT)

    def test_interface_diff_lipschitz(self, op):
        # .diff_lipschitz() always returns a float.
        self._skip_if_disabled()
        dL = op.diff_lipschitz
        assert isinstance(dL, pxt.Real)

    def test_interface2_diff_lipschitz(self, op):
        # .diff_lipschitz() can be manually set.
        # [and negative values are illegal.]
        self._skip_if_disabled()
        dL_orig = op.diff_lipschitz

        dL_new = self._random_array((1,)).item()
        op.diff_lipschitz = dL_new
        assert isinstance(op.diff_lipschitz, pxt.Real)
        with pxrt.EnforcePrecision(False):
            assert np.isclose(op.diff_lipschitz, dL_new)

        with pytest.raises(Exception):
            op.diff_lipschitz = -1

        # reset state for next tests
        op.diff_lipschitz = dL_orig

    def test_precCM_diff_lipschitz(self, op, width):
        self._skip_if_disabled()
        func = lambda: op.diff_lipschitz
        self._check_precCM_func(func, width)

    # We disable RuntimeWarnings which may arise due to NaNs. (See comment below.)
    @pytest.mark.filterwarnings("ignore::RuntimeWarning")
    def test_math_diff_lipschitz(
        self,
        op,
        xp,
        width,
        data_math_diff_lipschitz,
        _data_estimate_diff_lipschitz,
    ):
        # \norm{J(x) - J(y)}{F} \le dL * \norm{x - y}{2}
        # dL provided by `estimate_diff_lipschitz()`.
        self._skip_if_disabled()
        with pxrt.EnforcePrecision(False):
            dL = op.estimate_diff_lipschitz(**_data_estimate_diff_lipschitz["in_"])

            J = lambda _: op.jacobian(_).asarray(dtype=width.value).flatten()
            # .flatten() used to consistently compare jacobians via the L2 norm.
            # (Allows one to re-use this test for scalar-valued DiffMaps.)

            stats = []  # (x, y, condition success)
            data = xp.array(data_math_diff_lipschitz, dtype=width.value)
            for x, y in itertools.combinations(data, 2):
                lhs = pxm.norm(J(x) - J(y))
                rhs = dL * pxm.norm(x - y)
                if np.isnan(lhs):
                    # J() may return INFs, in which case `INF-INF=NaN` may arise above. less_equal()
                    # is not able to handle NaNs, so the former are overwritten by a sensible value
                    # in this context, i.e. 0.
                    lhs = 0
                success = ct.less_equal(lhs, rhs, as_dtype=width.value)
                stats.append((lhs, rhs, success))

            assert all(_[2] for _ in stats)


class DiffFuncT(FuncT, DiffMapT):
    # Class Properties --------------------------------------------------------
    base = pxa.DiffFunc
    interface = frozenset(FuncT.interface | DiffMapT.interface | {"grad"})
    disable_test = frozenset(FuncT.disable_test | DiffMapT.disable_test)

    # Fixtures ----------------------------------------------------------------
    @pytest.fixture
    def data_grad(self) -> DataLike:
        # override in subclass with 1D input/outputs of op.grad().
        # Arrays should be NumPy-only. (Internal machinery will transform to different
        # backend/precisions as needed.)
        raise NotImplementedError

    @pytest.fixture
    def _data_grad(self, data_grad, xp, width) -> DataLike:
        # Generate Cartesian product of inputs.
        # Do not override in subclass: for internal use only to test `op.grad()`.
        # Outputs are left unchanged: different tests should transform them as required.
        in_ = copy.deepcopy(data_grad["in_"])
        in_.update(arr=xp.array(in_["arr"], dtype=width.value))
        data = dict(
            in_=in_,
            out=data_grad["out"],
        )
        return data

    # Tests -------------------------------------------------------------------
    def test_value1D_grad(self, op, _data_grad):
        self._skip_if_disabled()
        self._check_value1D(op.grad, _data_grad)

    def test_valueND_grad(self, op, _data_grad):
        self._skip_if_disabled()
        self._check_valueND(op.grad, _data_grad)

    def test_backend_grad(self, op, _data_grad):
        self._skip_if_disabled()
        self._check_backend(op.grad, _data_grad)

    def test_prec_grad(self, op, _data_grad):
        self._skip_if_disabled()
        self._check_prec(op.grad, _data_grad)

    def test_precCM_grad(self, op, _data_grad):
        self._skip_if_disabled()
        self._check_precCM(op.grad, _data_grad)

    def test_transparent_grad(self, op, _data_grad):
        self._skip_if_disabled()
        self._check_no_side_effect(op.grad, _data_grad)

    def test_math1_grad(self, op, width, _data_grad):
        # .jacobian/.grad outputs are consistent.
        self._skip_if_disabled()
        arr = _data_grad["in_"]["arr"]
        with pxrt.EnforcePrecision(False):
            J = op.jacobian(arr).asarray(dtype=width.value)
            g = op.grad(arr)

            assert J.size == g.size
            assert ct.allclose(J.squeeze(), g, as_dtype=arr.dtype)

    def test_math2_grad(self, op, xp, width, _data_estimate_lipschitz):
        # f(x - \frac{1}{L} \grad_{f}(x)) <= f(x)
        # L provided by `estimate_lipschitz()`
        self._skip_if_disabled()
        with pxrt.EnforcePrecision(False):
            L = op.estimate_lipschitz(**_data_estimate_lipschitz["in_"])
            if np.isclose(L, 0):
                return  # trivially true since f(x) = cst
            elif np.isclose(L, np.inf):
                return  # trivially true since f(x) <= f(x)

            N_test, N_dim = 5, self._sanitize(op.dim, default=3)
            rhs = self._random_array((N_test, N_dim), xp=xp, width=width)
            lhs = rhs - op.grad(rhs) / L

            assert np.all(ct.less_equal(op.apply(lhs), op.apply(rhs), as_dtype=width.value))


class ProxFuncT(FuncT):
    # Class Properties --------------------------------------------------------
    base = pxa.ProxFunc
    interface = frozenset(
        FuncT.interface
        | {
            "prox",
            "fenchel_prox",
            "moreau_envelope",
        }
    )

    # Fixtures ----------------------------------------------------------------
    @pytest.fixture
    def _op_m(self, op) -> tuple[float, pxa.DiffFunc]:
        mu = 1.1
        return mu, op.moreau_envelope(mu)

    @pytest.fixture
    def data_prox(self) -> DataLike:
        # override in subclass with 1D input/outputs of op.prox().
        # Arrays should be NumPy-only. (Internal machinery will transform to different
        # backend/precisions as needed.)
        raise NotImplementedError

    @pytest.fixture
    def data_fenchel_prox(self, data_prox) -> DataLike:
        # override in subclass with 1D input/outptus of op.fenchel_prox().
        # Default value: inferred from data_prox().
        p_arr = data_prox["in_"]["arr"]
        p_tau = data_prox["in_"]["tau"]
        p_out = data_prox["out"]
        return dict(
            in_=dict(
                arr=p_arr / p_tau,
                sigma=1 / p_tau,
            ),
            out=(p_arr - p_out) / p_tau,
        )

    @pytest.fixture
    def _data_prox(self, data_prox, xp, width) -> DataLike:
        # Generate Cartesian product of inputs.
        # Do not override in subclass: for internal use only to test `op.prox()`.
        # Outputs are left unchanged: different tests should transform them as required.
        in_ = copy.deepcopy(data_prox["in_"])
        in_.update(arr=xp.array(in_["arr"], dtype=width.value))
        data = dict(
            in_=in_,
            out=data_prox["out"],
        )
        return data

    @pytest.fixture
    def _data_fenchel_prox(self, data_fenchel_prox, xp, width) -> DataLike:
        # Generate Cartesian product of inputs.
        # Do not override in subclass: for internal use only to test `op.fenchel_prox()`.
        # Outputs are left unchanged: different tests should transform them as required.
        in_ = copy.deepcopy(data_fenchel_prox["in_"])
        in_.update(arr=xp.array(in_["arr"], dtype=width.value))
        data = dict(
            in_=in_,
            out=data_fenchel_prox["out"],
        )
        return data

    # Tests -------------------------------------------------------------------
    def test_value1D_prox(self, op, _data_prox):
        self._skip_if_disabled()
        self._check_value1D(op.prox, _data_prox)

    def test_valueND_prox(self, op, _data_prox):
        self._skip_if_disabled()
        self._check_valueND(op.prox, _data_prox)

    def test_backend_prox(self, op, _data_prox):
        self._skip_if_disabled()
        self._check_backend(op.prox, _data_prox)

    def test_prec_prox(self, op, _data_prox):
        self._skip_if_disabled()
        self._check_prec(op.prox, _data_prox)

    def test_precCM_prox(self, op, _data_prox):
        self._skip_if_disabled()
        self._check_precCM(op.prox, _data_prox)

    def test_transparent_prox(self, op, _data_prox):
        self._skip_if_disabled()
        self._check_no_side_effect(op.prox, _data_prox)

    def test_math_prox(self, op, xp, width, _data_prox):
        # Ensure y = prox_{tau f}(x) minimizes:
        # 2\tau f(z) - \norm{z - x}{2}^{2}, for any z \in \bR^{N}
        self._skip_if_disabled()
        in_ = _data_prox["in_"]
        y = op.prox(**in_)

        N_test, N_dim = 5, y.shape[-1]
        x = self._random_array((N_test, N_dim), xp=xp, width=width) + in_["arr"]

        def g(x):
            a = 2 * in_["tau"] * op.apply(x)
            b = pxm.norm(in_["arr"] - x, axis=-1, keepdims=True) ** 2
            return a + b

        assert np.all(ct.less_equal(g(y), g(x), as_dtype=y.dtype))

    def test_value1D_fenchel_prox(self, op, _data_fenchel_prox):
        self._skip_if_disabled()
        self._check_value1D(op.fenchel_prox, _data_fenchel_prox)

    def test_valueND_fenchel_prox(self, op, _data_fenchel_prox):
        self._skip_if_disabled()
        self._check_valueND(op.fenchel_prox, _data_fenchel_prox)

    def test_backend_fenchel_prox(self, op, _data_fenchel_prox):
        self._skip_if_disabled()
        self._check_backend(op.fenchel_prox, _data_fenchel_prox)

    def test_prec_fenchel_prox(self, op, _data_fenchel_prox):
        self._skip_if_disabled()
        self._check_prec(op.fenchel_prox, _data_fenchel_prox)

    def test_precCM_fenchel_prox(self, op, _data_fenchel_prox):
        self._skip_if_disabled()
        self._check_precCM(op.fenchel_prox, _data_fenchel_prox)

    def test_transparent_fenchel_prox(self, op, _data_fenchel_prox):
        self._skip_if_disabled()
        self._check_no_side_effect(op.fenchel_prox, _data_fenchel_prox)

    def test_interface_moreau_envelope(self, _op_m):
        self._skip_if_disabled()
        _, op_m = _op_m
        self._check_has_interface(op_m, DiffFuncT)

    def test_math1_moreau_envelope(self, op, _op_m, _data_apply):
        # op_m.apply() lower-bounds op.apply()
        self._skip_if_disabled()
        _, op_m = _op_m
        arr = _data_apply["in_"]["arr"]
        lhs = op_m.apply(arr)
        rhs = op.apply(arr)

        assert ct.less_equal(lhs, rhs, as_dtype=rhs.dtype)

    def test_math2_moreau_envelope(self, op, _op_m, _data_apply):
        # op_m.grad(x) * mu = x - op.prox(x, mu)
        self._skip_if_disabled()
        mu, op_m = _op_m
        arr = _data_apply["in_"]["arr"]
        lhs = op_m.grad(arr) * mu
        rhs = arr - op.prox(arr, mu)

        assert self._metric(lhs, rhs, as_dtype=arr.dtype)


class ProxDiffFuncT(ProxFuncT, DiffFuncT):
    # Class Properties --------------------------------------------------------
    base = pxa.ProxDiffFunc
    interface = frozenset(ProxFuncT.interface | DiffFuncT.interface)


@pytest.mark.filterwarnings("ignore::pyxu.info.warning.DenseWarning")
class QuadraticFuncT(ProxDiffFuncT):
    # Class Properties --------------------------------------------------------
    base = pxa.QuadraticFunc
    interface = ProxDiffFuncT.interface

    @pytest.fixture
    def data_math_lipschitz(self, op):
        N_test, dim = 5, self._sanitize(op.dim, 3)
        return self._random_array((N_test, dim), seed=5)

    @pytest.fixture
    def data_math_diff_lipschitz(self, op):
        N_test, dim = 5, self._sanitize(op.dim, 3)
        return self._random_array((N_test, dim), seed=5)

    @pytest.fixture(params=["svd", "trace"])  # supported methods
    def _data_estimate_diff_lipschitz(self, xp, width, _gpu, request) -> DataLike:
        data = dict(
            in_=dict(
                method=request.param,
                gpu=_gpu,
                xp=xp,
                dtype=width.value,
            )
        )
        return data

    # Tests -------------------------------------------------------------------
    # [fenchel_]prox() use CG internally.
    # To avoid CG convergence issues, correctness is assesed at lowest precision only.
    def test_value1D_prox(self, op, _data_prox):
        self._skip_if_disabled()
        self._check_value1D(op.prox, _data_prox, dtype=np.dtype(np.half))

    def test_valueND_prox(self, op, _data_prox):
        self._skip_if_disabled()
        self._check_valueND(op.prox, _data_prox, dtype=np.dtype(np.half))

    def test_value1D_fenchel_prox(self, op, _data_fenchel_prox):
        self._skip_if_disabled()
        self._check_value1D(op.fenchel_prox, _data_fenchel_prox, dtype=np.dtype(np.half))

    def test_valueND_fenchel_prox(self, op, _data_fenchel_prox):
        self._skip_if_disabled()
        self._check_valueND(op.fenchel_prox, _data_fenchel_prox, dtype=np.dtype(np.half))

    def test_math_quad_spec(self, op, xp, width):
        # f(0) == t
        # \grad_{f}(0) == c
        # J_{f}(e_{i}) - c == Q_{i}
        self._skip_if_disabled()
        Q, c, t = op._quad_spec()
        Q = Q.asarray(xp=xp, dtype=width.value)
        c = c.asarray(xp=xp, dtype=width.value)

        # f(0) == t
        x = xp.zeros(op.dim, dtype=width.value)
        out = op.apply(x)
        assert self._metric(out, t, as_dtype=width.value)

        # \grad_{f}(0) == c
        out = op.grad(x)
        assert self._metric(out, c, as_dtype=width.value)

        # J_{f}(e_{i}) - c == Q_{i}
        x = xp.eye(op.dim)
        out = op.grad(x) - c
        assert self._metric(out.T, Q, as_dtype=width.value)

    def test_math2_diff_lipschitz(self, op, _data_estimate_diff_lipschitz, _gpu):
        # op.estimate_diff_lipschitz(svd) <= op.estimate_diff_lipschitz(trace)
        self._skip_if_disabled()
        kwargs = _data_estimate_diff_lipschitz["in_"].copy()
        kwargs.pop("method")  # over-written below ...
        dtype = kwargs["dtype"]

        dL_ub = op.estimate_diff_lipschitz(method="trace", **kwargs)
        dL_opt = ct.flaky(  # might raise exception in GPU-mode
            op.estimate_diff_lipschitz,
            args=dict(method="svd", **kwargs),
            condition=_gpu is True,
            reason="svdvals() sparse-evaled via CuPy flaky.",
        )

        try:
            assert ct.less_equal(dL_opt, dL_ub, as_dtype=dtype).item()
        except AssertionError:
            if _gpu is True:
                # even if dL_opt computed, not guaranteed to be correct in GPU-mode
                pytest.xfail("svdvals() sparse-evaled via CuPy flaky.")
            else:
                raise

    def test_math3_diff_lipschitz(self, op, _data_estimate_diff_lipschitz, _gpu):
        # op.estimate_diff_lipschitz(svd) computes the optimal diff-Lipschitz constant.
        self._skip_if_disabled()
        kwargs = _data_estimate_diff_lipschitz["in_"].copy()
        kwargs.pop("method")  # over-written below ...
        dtype = kwargs["dtype"]

        dL_svds = ct.flaky(  # might raise an assertion in GPU-mode
            func=op.estimate_diff_lipschitz,
            args=dict(method="svd", **kwargs),
            condition=_gpu is True,
            reason="svdvals() sparse-evaled via CuPy flaky.",
        )

        Q, _, _ = op._quad_spec()
        QL_svds = ct.flaky(  # might raise an assertion in GPU-mode
            func=Q.estimate_lipschitz,
            args=dict(method="svd", **kwargs),
            condition=_gpu is True,
            reason="svdvals() sparse-evaled via CuPy flaky.",
        )

        # Comparison is done with user-specified metric since operator may compute `dL_svds` via an
        # approximate `.apply()' method.
        try:
            cast = lambda x: np.array([x], dtype=dtype)
            assert self._metric(cast(dL_svds), cast(QL_svds), as_dtype=dtype)
        except AssertionError:
            if _gpu is True:
                # even if dL_opt computed, not guaranteed to be correct in GPU-mode
                pytest.xfail("svdvals() sparse-evaled via CuPy flaky.")
            else:
                raise


@pytest.mark.filterwarnings("ignore::pyxu.info.warning.DenseWarning")
class LinOpT(DiffMapT):
    # Class Properties --------------------------------------------------------
    base = pxa.LinOp
    interface = frozenset(
        DiffMapT.interface
        | {
            "adjoint",
            "asarray",
            "cogram",
            "dagger",
            "from_array",
            "gram",
            "pinv",
            "svdvals",
            "T",
            "to_sciop",
        }
    )

    # Internal helpers --------------------------------------------------------
    def _skip_unless_NUMPY_CUPY(self, ndi):
        N = pxd.NDArrayInfo
        if ndi not in {N.NUMPY, N.CUPY}:
            pytest.skip("Only NUMPY/CUPY backends supported.")

    @classmethod
    def _check_value1D_vals(cls, func, kwargs, ground_truth):
        N = pxd.NDArrayInfo
        xp = N.from_flag(kwargs["gpu"]).module()
        if kwargs["gpu"]:
            ground_truth = xp.array(ground_truth)

        k = kwargs["k"]
        out = func(**kwargs)
        idx = xp.argsort(xp.abs(out))
        assert out.size == k  # obtain N_vals asked for
        assert ct.allclose(out[idx], out, out.dtype)  # sorted in ascending magnitude

        # and output is correct (in magnitude)
        which = kwargs["which"]
        idx_gt = xp.argsort(xp.abs(ground_truth))
        if which == "SM":
            out = out[idx][:k]
            gt = ground_truth[idx_gt][:k]
        else:  # LM
            out = out[idx][-k:]
            gt = ground_truth[idx_gt][-k:]

        # When seeking the smallest-magnitude singular values via svdvals(), the iterative algorithm
        # used converges to values close to the ground-truth (GT).
        # However, for 0-valued singular vectors, the relative error between converged solution and
        # GT is sometimes slightly higher than the relative tolerance set for FP32/FP64 to consider
        # them identical.
        # We therefore assess svdvals() outputs at FP32-precision only.
        # This precision is enough to query an operator's spectrum for further diagnostics.
        assert cls._metric(xp.abs(out), xp.abs(gt), as_dtype=pxrt.Width.SINGLE.value)

    @staticmethod
    def _check_backend_vals(func, _gpu):
        N = pxd.NDArrayInfo
        out = func(k=1, gpu=_gpu)
        assert N.from_obj(out) == N.from_flag(_gpu)

    # Fixtures ----------------------------------------------------------------
    @pytest.fixture
    def data_adjoint(self, op) -> DataLike:
        # override in subclass with 1D input/outputs of op.adjoint().
        # Arrays should be NumPy-only. (Internal machinery will transform to different
        # backend/precisions as needed.)
        #
        # Default implementation just tests A(0) = 0. Subsequent math tests ensure .adjoint() output
        # is consistent with .apply().
        return dict(
            in_=dict(arr=np.zeros(op.codim)),
            out=np.zeros(op.dim),
        )

    @pytest.fixture
    def _data_adjoint(self, data_adjoint, xp, width) -> DataLike:
        # Generate Cartesian product of inputs.
        # Do not override in subclass: for internal use only to test `op.adjoint()`.
        # Outputs are left unchanged: different tests should transform them as required.
        in_ = copy.deepcopy(data_adjoint["in_"])
        in_.update(arr=xp.array(in_["arr"], dtype=width.value))
        data = dict(
            in_=in_,
            out=data_adjoint["out"],
        )
        return data

    @pytest.fixture
    def data_math_lipschitz(self, op) -> cabc.Collection[np.ndarray]:
        N_test = 5
        return self._random_array((N_test, op.dim))

    @pytest.fixture
    def _data_estimate_lipschitz(self, xp, width, _gpu) -> DataLike:
        data = dict(
            in_=dict(
                # method=...  # defined in tests as needed
                gpu=_gpu,
                xp=xp,
                dtype=width.value,
            )
        )
        return data

    @pytest.fixture
    def data_math_diff_lipschitz(self, op) -> cabc.Collection[np.ndarray]:
        N_test = 5
        return self._random_array((N_test, op.dim))

    @pytest.fixture
    def _op_svd(self, op) -> np.ndarray:
        # compute all singular values, sorted in ascending order.
        D = np.linalg.svd(
            op.asarray(),
            full_matrices=False,
            compute_uv=False,
        )
        return np.sort(D)

    @pytest.fixture(params=[1, 2])
    def _damp(self, request) -> float:
        # candidate dampening factors for .pinv() & .dagger()
        # We do not test damp=0 since ill-conditioning of some operators would require a lot of
        # manual _damp-correcting to work.
        return request.param

    @pytest.fixture
    def data_pinv(self, op, _damp, data_apply) -> DataLike:
        # override in subclass with 1D input/outputs of op.pinv().
        # Arrays should be NumPy-only. (Internal machinery will transform to different
        # backend/precisions as needed.)
        #
        # Default implementation: auto-computes pinv() at output points specified to test op.apply().

        # Safe implementation of --------------------------
        #   A = op.gram().asarray(xp=np, dtype=pxrt.Width.DOUBLE.value)
        B = op.asarray(xp=np, dtype=pxrt.Width.DOUBLE.value)
        A = B.T @ B
        # -------------------------------------------------
        for i in range(op.dim):
            A[i, i] += _damp

        arr = data_apply["out"]
        out, *_ = splinalg.lstsq(A, B.T @ arr)
        data = dict(
            in_=dict(
                arr=arr,
                damp=_damp,
                kwargs_init=dict(),
                kwargs_fit=dict(),
            ),
            out=out,
        )
        return data

    @pytest.fixture
    def _data_pinv(self, data_pinv, xp, width) -> DataLike:
        # Generate Cartesian product of inputs.
        # Do not override in subclass: for internal use only to test `op.pinv()`.
        # Outputs are left unchanged: different tests should transform them as required.
        in_ = copy.deepcopy(data_pinv["in_"])
        in_.update(arr=xp.array(in_["arr"], dtype=width.value))
        data = dict(
            in_=in_,
            out=data_pinv["out"],
        )
        return data

    @pytest.fixture
    def _op_dagger(self, op, _damp) -> pxa.LinOp:
        op_d = op.dagger(damp=_damp)
        return op_d

    @pytest.fixture
    def _data_apply_dagger(self, _data_pinv) -> DataLike:
        # Do not override in subclass: for internal use only to test `op.dagger().apply()`.
        #
        # Clone of _data_pinv, with all arguments unsupported by .apply() removed.
        data = copy.deepcopy(_data_pinv)
        data["in_"] = dict(arr=data["in_"]["arr"])
        return data

    @pytest.fixture(
        params=itertools.product(
            pxd.supported_array_modules(),
            pxrt.Width,
        )
    )
    def _op_array(self, op, xp, width, request) -> pxt.NDArray:
        # Ground-truth array which should be returned by .asarray()

        # op() only defined for specifid xp/width combos.
        # Idea: compute .asarray() using backend/precision supported by op(), then cast to
        # user-desired backend/precision.
        A_gt = xp.zeros((op.codim, op.dim), dtype=width.value)
        for i in range(op.dim):
            e = xp.zeros((op.dim,), dtype=width.value)
            e[i] = 1
            with pxrt.EnforcePrecision(False):
                A_gt[:, i] = op.apply(e)

        N = pxd.NDArrayInfo
        xp_, width_ = request.param
        if N.from_obj(A_gt) == N.CUPY:
            A_gt = A_gt.get()
        return xp_.array(A_gt, dtype=width_.value)

    @pytest.fixture
    def _op_sciop(self, op, _gpu, width) -> spsl.LinearOperator:
        A = op.to_sciop(dtype=width.value, gpu=_gpu)
        return A

    @pytest.fixture(
        params=[
            "matvec",
            "matmat",
            "rmatvec",
            "rmatmat",
        ]
    )
    def _data_to_sciop(self, op, ndi, xp, width, request) -> DataLike:
        # Do not override in subclass: for internal use only to test `op.to_sciop()`.
        self._skip_unless_NUMPY_CUPY(ndi)

        N_test = 7
        f = lambda _: self._random_array(_, xp=xp, width=width)
        op_array = op.asarray(xp=xp, dtype=width.value)
        mode = request.param
        if mode == "matvec":
            arr = f((op.dim,))
            out_gt = op_array @ arr
            var = "x"
        elif mode == "matmat":
            arr = f((op.dim, N_test))
            out_gt = op_array @ arr
            var = "X"
        elif mode == "rmatvec":
            arr = f((op.codim,))
            out_gt = op_array.T @ arr
            var = "x"
        elif mode == "rmatmat":
            arr = f((op.codim, N_test))
            out_gt = op_array.T @ arr
            var = "X"
        return dict(
            in_={var: arr},
            out=out_gt,
            mode=mode,  # for test_xxx_sciop()
        )

    # Tests -------------------------------------------------------------------
    def test_value1D_adjoint(self, op, _data_adjoint):
        self._skip_if_disabled()
        self._check_value1D(op.adjoint, _data_adjoint)

    def test_valueND_adjoint(self, op, _data_adjoint):
        self._skip_if_disabled()
        self._check_valueND(op.adjoint, _data_adjoint)

    def test_backend_adjoint(self, op, _data_adjoint):
        self._skip_if_disabled()
        self._check_backend(op.adjoint, _data_adjoint)

    def test_prec_adjoint(self, op, _data_adjoint):
        self._skip_if_disabled()
        self._check_prec(op.adjoint, _data_adjoint)

    def test_precCM_adjoint(self, op, _data_adjoint):
        self._skip_if_disabled()
        self._check_precCM(op.adjoint, _data_adjoint)

    def test_transparent_adjoint(self, op, _data_adjoint):
        self._skip_if_disabled()
        self._check_no_side_effect(op.adjoint, _data_adjoint)

    def test_math_adjoint(self, op, xp, width):
        # <op.adjoint(x), y> = <x, op.apply(y)>
        self._skip_if_disabled()
        N = 20
        x = self._random_array((N, op.codim), xp=xp, width=width)
        y = self._random_array((N, op.dim), xp=xp, width=width)

        ip = lambda a, b: (a * b).sum(axis=-1)  # (N, Q) * (N, Q) -> (N,)
        with pxrt.EnforcePrecision(False):
            lhs = ip(op.adjoint(x), y)
            rhs = ip(x, op.apply(y))

        assert self._metric(lhs, rhs, as_dtype=width.value)

    def test_math2_lipschitz(self, op, _data_estimate_lipschitz, _gpu):
        # op.estimate_lipschitz(svd) <= op.estimate_lipschitz(trace)
        self._skip_if_disabled()
        kwargs = _data_estimate_lipschitz["in_"]
        dtype = kwargs["dtype"]

        L_ub = op.estimate_lipschitz(method="trace", **kwargs)
        L_opt = ct.flaky(  # might raise exception in GPU-mode
            op.estimate_lipschitz,
            args=dict(method="svd", **kwargs),
            condition=_gpu is True,
            reason="svdvals() sparse-evaled via CuPy flaky.",
        )

        try:
            assert ct.less_equal(L_opt, L_ub, as_dtype=dtype).item()
        except AssertionError:
            if _gpu is True:
                # even if L_opt computed, not guaranteed to be correct in GPU-mode
                pytest.xfail("svdvals() sparse-evaled via CuPy flaky.")
            else:
                raise

    def test_math3_lipschitz(self, op, _data_estimate_lipschitz, _op_svd, _gpu):
        # op.estimate_lipschitz(svd) computes the optimal Lipschitz constant.
        self._skip_if_disabled()
        kwargs = _data_estimate_lipschitz["in_"]
        dtype = kwargs["dtype"]

        L_svds = ct.flaky(  # might raise an assertion in GPU-mode
            func=op.estimate_lipschitz,
            args=dict(method="svd", **kwargs),
            condition=_gpu is True,
            reason="svdvals() sparse-evaled via CuPy flaky.",
        )
        # Comparison is done with user-specified metric since operator may compute `L_svds` via an
        # approximate `.apply()' method.
        try:
            cast = lambda x: np.array([x], dtype=dtype)
            assert self._metric(cast(L_svds), cast(_op_svd.max()), as_dtype=dtype)
        except AssertionError:
            if _gpu is True:
                # even if L_opt computed, not guaranteed to be correct in GPU-mode
                pytest.xfail("svdvals() sparse-evaled via CuPy flaky.")
            else:
                raise

    def test_interface_jacobian(self, op, _data_apply):
        self._skip_if_disabled()
        arr = _data_apply["in_"]["arr"]
        J = op.jacobian(arr)
        assert J is op

    @pytest.mark.parametrize("k", [1, 2])
    @pytest.mark.parametrize("which", ["LM", "SM"])
    def test_value1D_svdvals(self, op, ndi, _gpu, _op_svd, k, which):
        self._skip_if_disabled()
        self._skip_unless_NUMPY_CUPY(ndi)
        gpu_reason = dict(
            LM="`which=LM` sparse-evaled via CuPy flaky.",
            SM="`which=SM` unsupported by CuPy",
        )
        ct.flaky(
            func=self._check_value1D_vals,
            args=dict(
                func=op.svdvals,
                kwargs=dict(k=k, which=which, gpu=_gpu),
                ground_truth=_op_svd,
            ),
            condition=_gpu is True,
            reason=gpu_reason[which],
        )

    def test_backend_svdvals(self, op, ndi, _gpu):
        self._skip_if_disabled()
        self._skip_unless_NUMPY_CUPY(ndi)
        ct.flaky(
            func=self._check_backend_vals,
            args=dict(func=op.svdvals, _gpu=_gpu),
            condition=_gpu is True,
            reason="svdvals() sparse-evaled via CuPy flaky.",
        )

    def test_precCM_svdvals(self, op, ndi, _gpu, width):
        self._skip_if_disabled()
        self._skip_unless_NUMPY_CUPY(ndi)
        data = dict(in_=dict(k=1, gpu=_gpu))
        ct.flaky(
            func=self._check_precCM,
            args=dict(func=op.svdvals, data=data, widths=(width,)),
            condition=_gpu is True,
            reason="svdvals() sparse-evaled via CuPy flaky.",
        )

    def test_value1D_pinv(self, op, _data_pinv):
        # To avoid CG convergence issues, correctness is assesed at lowest precision only.
        self._skip_if_disabled()
        self._check_value1D(op.pinv, _data_pinv, dtype=np.dtype(np.half))

    def test_valueND_pinv(self, op, _data_pinv):
        # To avoid CG convergence issues, correctness is assesed at lowest precision only.
        self._skip_if_disabled()
        self._check_valueND(op.pinv, _data_pinv, dtype=np.dtype(np.half))

    def test_backend_pinv(self, op, _data_pinv):
        self._skip_if_disabled()
        self._check_backend(op.pinv, _data_pinv)

    def test_prec_pinv(self, op, _data_pinv):
        self._skip_if_disabled()
        self._check_prec(op.pinv, _data_pinv)

    def test_precCM_pinv(self, op, _data_pinv):
        self._skip_if_disabled()
        self._check_precCM(op.pinv, _data_pinv)

    def test_interface_dagger(self, _op_dagger):
        self._skip_if_disabled()
        self._check_has_interface(_op_dagger, LinOpT)

    def test_value1D_call_dagger(self, _op_dagger, _data_apply_dagger):
        # To avoid CG convergence issues, correctness is assesed at lowest precision only.
        self._skip_if_disabled()
        self._check_value1D(_op_dagger.__call__, _data_apply_dagger, dtype=np.dtype(np.half))

    def test_valueND_call_dagger(self, _op_dagger, _data_apply_dagger):
        # To avoid CG convergence issues, correctness is assesed at lowest precision only.
        self._skip_if_disabled()
        self._check_valueND(_op_dagger.__call__, _data_apply_dagger, dtype=np.dtype(np.half))

    def test_backend_call_dagger(self, _op_dagger, _data_apply_dagger):
        self._skip_if_disabled()
        self._check_backend(_op_dagger.__call__, _data_apply_dagger)

    def test_prec_call_dagger(self, _op_dagger, _data_apply_dagger):
        self._skip_if_disabled()
        self._check_prec(_op_dagger.__call__, _data_apply_dagger)

    def test_precCM_call_dagger(self, _op_dagger, _data_apply_dagger):
        self._skip_if_disabled()
        self._check_precCM(_op_dagger.__call__, _data_apply_dagger)

    def test_value1D_apply_dagger(self, _op_dagger, _data_apply_dagger):
        # To avoid CG convergence issues, correctness is assesed at lowest precision only.
        self._skip_if_disabled()
        self._check_value1D(_op_dagger.apply, _data_apply_dagger, dtype=np.dtype(np.half))

    def test_valueND_apply_dagger(self, _op_dagger, _data_apply_dagger):
        # To avoid CG convergence issues, correctness is assesed at lowest precision only.
        self._skip_if_disabled()
        self._check_valueND(_op_dagger.apply, _data_apply_dagger, dtype=np.dtype(np.half))

    def test_backend_apply_dagger(self, _op_dagger, _data_apply_dagger):
        self._skip_if_disabled()
        self._check_backend(_op_dagger.apply, _data_apply_dagger)

    def test_prec_apply_dagger(self, _op_dagger, _data_apply_dagger):
        self._skip_if_disabled()
        self._check_prec(_op_dagger.apply, _data_apply_dagger)

    def test_precCM_apply_dagger(self, _op_dagger, _data_apply_dagger):
        self._skip_if_disabled()
        self._check_precCM(_op_dagger.apply, _data_apply_dagger)

    def test_interface_gram(self, op):
        self._skip_if_disabled()
        if op.dim > 1:
            klass = SelfAdjointOpT
        else:
            klass = LinFuncT
        self._check_has_interface(op.gram(), klass)

    def test_math_gram(self, op, xp, width):
        # op_g.apply == op_g.adjoint == adjoint \comp apply
        self._skip_if_disabled()
        op_g = op.gram()
        x = self._random_array((30, op.dim), xp=xp, width=width)

        with pxrt.EnforcePrecision(False):
            assert self._metric(op_g.apply(x), op_g.adjoint(x), as_dtype=width.value)
            assert self._metric(op_g.apply(x), op.adjoint(op.apply(x)), as_dtype=width.value)

    def test_interface_cogram(self, op):
        self._skip_if_disabled()
        if op.codim > 1:
            klass = SelfAdjointOpT
        else:
            klass = LinFuncT
        self._check_has_interface(op.cogram(), klass)

    def test_math_cogram(self, op, xp, width):
        # op_cg.apply == op_cg.adjoint == apply \comp adjoint
        self._skip_if_disabled()
        op_cg = op.cogram()
        x = self._random_array((30, op.codim), xp=xp, width=width)

        with pxrt.EnforcePrecision(False):
            assert self._metric(op_cg.apply(x), op_cg.adjoint(x), as_dtype=width.value)
            assert self._metric(op_cg.apply(x), op.apply(op.adjoint(x)), as_dtype=width.value)

    def test_value_asarray(self, op, _op_array):
        self._skip_if_disabled()
        xp = pxu.get_array_module(_op_array)
        dtype = _op_array.dtype

        A = op.asarray(xp=xp, dtype=dtype)
        assert A.shape == _op_array.shape
        assert self._metric(_op_array.T, A.T, as_dtype=dtype)

    def test_backend_asarray(self, op, _op_array):
        self._skip_if_disabled()
        xp = pxu.get_array_module(_op_array)
        dtype = _op_array.dtype

        A = op.asarray(xp=xp, dtype=dtype)
        assert pxu.get_array_module(A) == xp

    def test_prec_asarray(self, op, _op_array):
        self._skip_if_disabled()
        xp = pxu.get_array_module(_op_array)
        dtype = _op_array.dtype

        A = op.asarray(xp=xp, dtype=dtype)
        assert A.dtype == dtype

    def test_value_to_sciop(self, _op_sciop, _data_to_sciop):
        self._skip_if_disabled()
        func = getattr(_op_sciop, _data_to_sciop["mode"])
        out = func(**_data_to_sciop["in_"])
        out_gt = _data_to_sciop["out"]
        assert out.shape == out_gt.shape
        assert self._metric(out, out_gt, as_dtype=out_gt.dtype)

    def test_backend_to_sciop(self, _op_sciop, _data_to_sciop):
        self._skip_if_disabled()
        func = getattr(_op_sciop, _data_to_sciop["mode"])
        out = func(**_data_to_sciop["in_"])
        out_gt = _data_to_sciop["out"]
        assert pxu.get_array_module(out) == pxu.get_array_module(out_gt)

    def test_prec_to_sciop(self, _op_sciop, _data_to_sciop):
        self._skip_if_disabled()
        func = getattr(_op_sciop, _data_to_sciop["mode"])
        out = func(**_data_to_sciop["in_"])
        out_gt = _data_to_sciop["out"]
        assert out.dtype == out_gt.dtype


class LinFuncT(ProxDiffFuncT, LinOpT):
    # Class Properties --------------------------------------------------------
    base = pxa.LinFunc
    interface = frozenset(ProxDiffFuncT.interface | LinOpT.interface)
    disable_test = frozenset(ProxDiffFuncT.disable_test | LinOpT.disable_test)

    # Fixtures ----------------------------------------------------------------
    @pytest.fixture
    def data_grad(self, op) -> DataLike:
        # We know for linfuncs that op.grad(x) = op.asarray()
        x = self._random_array((op.dim,))
        y = op.asarray()
        return dict(
            in_=dict(arr=x),
            out=y,
        )

    @pytest.fixture
    def data_adjoint(self, data_grad) -> DataLike:
        # We know for linfuncs that op.adj(x) = op.grad(x) * x
        x = self._random_array((1,))
        y = x * data_grad["out"]
        return dict(
            in_=dict(arr=x),
            out=y,
        )

    @pytest.fixture
    def data_prox(self, data_grad) -> DataLike:
        # We know for linfuncs that op.prox(x, tau) = x - op.grad(x) * tau
        x = data_grad["in_"]["arr"].copy()
        g = data_grad["out"]
        tau = np.abs(self._random_array((1,)))[0]
        y = x - tau * g
        return dict(
            in_=dict(
                arr=x,
                tau=tau,
            ),
            out=y,
        )

    # Tests -------------------------------------------------------------------
    @pytest.mark.parametrize("k", [1])  # override: only `k=1` feasible for LinFuncs
    @pytest.mark.parametrize("which", ["SM", "LM"])
    def test_value1D_svdvals(self, op, ndi, _gpu, _op_svd, k, which):
        self._skip_if_disabled()
        super().test_value1D_svdvals(op, ndi, _gpu, _op_svd, k, which)

    @pytest.mark.skip("Notion of loss undefined for LinFuncs.")
    def test_interface_asloss(self, op, xp, width):
        super().test_interface_asloss(op, xp, width)


class SquareOpT(LinOpT):
    # Class Properties --------------------------------------------------------
    base = pxa.SquareOp
    interface = frozenset(LinOpT.interface | {"trace"})

    # Fixtures ----------------------------------------------------------------
    @pytest.fixture
    def _op_trace(self, op) -> float:
        # True trace
        tr = op.asarray().trace()
        return float(tr)

    @pytest.fixture
    def _data_trace(self, xp, width) -> DataLike:
        # Generate Cartesian product of inputs.
        # Do not override in subclass: for internal use only to test `op.trace()`.
        data = dict(
            in_=dict(
                method="explicit",  # hutchpp() variant is time-consuming: tested in test_linalg.py
                xp=xp,
                dtype=width.value,
            )
        )
        return data

    # Tests -------------------------------------------------------------------
    def test_square(self, op):
        self._skip_if_disabled()
        assert op.dim == op.codim

    def test_interface_trace(self, op, _data_trace):
        tr = op.trace(**_data_trace["in_"])
        assert isinstance(tr, pxt.Real)

    def test_value_trace(self, op, _data_trace, _op_trace, width):
        # tr(op) == op.trace(method=explicit)
        tr = op.trace(**_data_trace["in_"])
        assert ct.allclose(tr, _op_trace, as_dtype=width.value)

    def test_precCM_trace(self, op, _data_trace, width):
        self._skip_if_disabled()
        func = lambda: op.trace(**_data_trace["in_"])
        self._check_precCM_func(func, width)


class NormalOpT(SquareOpT):
    # Class Properties --------------------------------------------------------
    base = pxa.NormalOp

    # Tests -------------------------------------------------------------------
    def test_math_normality(self, op, xp, width):
        # AA^{*} = A^{*}A
        self._skip_if_disabled()
        N = 20
        x = self._random_array((N, op.dim), xp=xp, width=width)

        lhs = op.apply(op.adjoint(x))
        rhs = op.adjoint(op.apply(x))
        assert self._metric(lhs, rhs, as_dtype=width.value)


class UnitOpT(NormalOpT):
    # Class Properties --------------------------------------------------------
    base = pxa.UnitOp

    # Internal helpers --------------------------------------------------------
    @classmethod
    def _check_identity(cls, operator, xp, width):
        x = cls._random_array((30, operator.dim), xp=xp, width=width)
        assert ct.allclose(operator.apply(x), x, as_dtype=width.value)
        assert ct.allclose(operator.adjoint(x), x, as_dtype=width.value)

    # Fixtures ----------------------------------------------------------------
    @pytest.fixture
    def _op_svd(self, op) -> np.ndarray:
        D = np.ones(op.dim)
        return D

    # Tests -------------------------------------------------------------------
    def test_math_gram(self, op, xp, width):
        # op_g == I
        self._skip_if_disabled()
        self._check_identity(op.gram(), xp, width)

    def test_math_cogram(self, op, xp, width):
        # op_cg == I
        self._skip_if_disabled()
        self._check_identity(op.cogram(), xp, width)

    def test_math_norm(self, op, xp, width):
        # \norm{U x} = \norm{U^{*} x} = \norm{x}
        self._skip_if_disabled()
        N = 20
        x = self._random_array((N, op.dim), xp=xp, width=width)

        lhs1 = pxm.norm(op.apply(x), axis=-1)
        lhs2 = pxm.norm(op.adjoint(x), axis=-1)
        rhs = pxm.norm(x, axis=-1)

        assert self._metric(lhs1, lhs2, as_dtype=width.value)
        assert self._metric(lhs1, rhs, as_dtype=width.value)


class SelfAdjointOpT(NormalOpT):
    # Class Properties --------------------------------------------------------
    base = pxa.SelfAdjointOp

    # Tests -------------------------------------------------------------------
    def test_math_selfadjoint(self, op, xp, width):
        # A = A^{*}
        self._skip_if_disabled()
        N = 20
        x = self._random_array((N, op.dim), xp=xp, width=width)

        lhs = op.apply(x)
        rhs = op.adjoint(x)
        assert ct.allclose(lhs, rhs, as_dtype=width.value)


class PosDefOpT(SelfAdjointOpT):
    # Class Properties --------------------------------------------------------
    base = pxa.PosDefOp

    # Tests -------------------------------------------------------------------
    def test_math_posdef(self, op, xp, width):
        # <Ax,x> > 0
        self._skip_if_disabled()
        N = 20
        x = self._random_array((N, op.dim), xp=xp, width=width)

        ip = lambda a, b: (a * b).sum(axis=-1)  # (N, Q) * (N, Q) -> (N,)
        assert np.all(ct.less_equal(0, ip(op.apply(x), x), as_dtype=width.value))


class ProjOpT(SquareOpT):
    # Class Properties --------------------------------------------------------
    base = pxa.ProjOp

    # Fixtures ----------------------------------------------------------------
    def test_math_idempotent(self, op, xp, width):
        # op.apply = op.apply^2
        self._skip_if_disabled()
        N = 30
        x = self._random_array((N, op.dim), xp=xp, width=width)
        y = op.apply(x)
        z = op.apply(y)

        assert self._metric(y, z, as_dtype=width.value)


class OrthProjOpT(ProjOpT, SelfAdjointOpT):
    # Class Properties --------------------------------------------------------
    base = pxa.OrthProjOp
    interface = frozenset(ProjOpT.interface | SelfAdjointOpT.interface)
    disable_test = frozenset(ProjOpT.disable_test | SelfAdjointOpT.disable_test)



import itertools

import numpy as np
import pytest

import pyxu.info.deps as pxd
import pyxu.operator as pxo
import pyxu.runtime as pxrt
import pyxu.util as pxu
import pyxu_tests.operator.conftest as conftest


# We disable PrecisionWarnings since DiagonalOp() is not precision-agnostic, but the outputs
# computed must still be valid.
@pytest.mark.filterwarnings("ignore::pyxu.info.warning.PrecisionWarning")
class TestDiagonalOp(conftest.PosDefOpT):
    @pytest.fixture
    def dim(self):
        return 20

    @pytest.fixture
    def data_shape(self, dim):
        return (dim, dim)

    @pytest.fixture(params=range(5))
    def vec(self, dim, request):
        v = self._random_array((dim,), seed=request.param)
        return v

    @pytest.fixture(
        params=itertools.product(
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def spec(self, vec, request):
        ndi = request.param[0]
        self._skip_if_unsupported(ndi)
        xp = ndi.module()
        width = request.param[1]

        vec = xp.array(vec, dtype=width.value)
        op = pxo.DiagonalOp(vec=vec)
        return op, *request.param

    @pytest.fixture
    def data_apply(self, vec):
        arr = 15 + np.arange(vec.size)
        out = vec * arr
        return dict(
            in_=dict(arr=arr),
            out=out,
        )

    def test_math_posdef(self, op, xp, width, vec):
        if np.any(pxu.compute(vec < 0)):
            pytest.skip("disabled since operator is not positive-definite.")
        else:
            super().test_math_posdef(op, xp, width)



import itertools

import numpy as np
import pytest

import pyxu.info.deps as pxd
import pyxu.operator as pxo
import pyxu.runtime as pxrt
import pyxu.util as pxu
import pyxu_tests.operator.conftest as conftest
import pyxu_tests.operator.linop.fft.conftest_nufft as conftest_nufft


class TestNUFFT2(conftest_nufft.NUFFT_Mixin, conftest.LinOpT):
    # (Extra) Fixtures which parametrize operator -----------------------------
    @pytest.fixture(params=[1, 13])
    def transform_x(self, transform_dimension, request) -> np.ndarray:
        # (M, D) D-dimensional sample points :math:`\mathbf{x}_{j} \in [-\pi, \pi)^{D}`.
        rng = np.random.default_rng(0)
        _x = rng.normal(size=(request.param, transform_dimension))
        x = np.fmod(_x, 2 * np.pi)
        return x

    @pytest.fixture(params=[2, 10])
    def transform_N(self, transform_dimension, request) -> tuple[int]:
        # (D,) mesh size in each dimension :math:`(N_1, \ldots, N_{D})`.
        rng = np.random.default_rng(1)
        N = rng.integers(
            low=1,
            high=request.param,
            size=(transform_dimension,),
            endpoint=True,
        )
        return tuple(N)

    @pytest.fixture
    def _transform_cArray(
        self,
        transform_x,
        transform_N,
        transform_dimension,
        transform_sign,
        transform_modeord,
    ) -> np.ndarray:
        # Ground-truth LinOp A: \bC^{N.prod()} -> \bC^{M} which encodes the type-2 transform.
        mesh = np.stack(  # (D, N1, ..., Nd)
            np.meshgrid(
                *[np.arange(-(n // 2), (n - 1) // 2 + 1) for n in transform_N],
                indexing="ij",
            ),
            axis=0,
        )
        if transform_modeord == 1:  # FFT order
            mesh = np.fft.ifftshift(mesh, axes=-(1 + np.arange(len(transform_N))))
        mesh = mesh.reshape((transform_dimension, -1))  # (D, N.prod())

        A = np.exp(1j * transform_sign * mesh.T @ transform_x.T)  # (N.prod(), M)
        return A.T  # (M, N.prod())

    # Fixtures from conftest.LinOpT -------------------------------------------
    @pytest.fixture(
        params=itertools.product(
            [
                pxd.NDArrayInfo.NUMPY,
                pxd.NDArrayInfo.DASK,
            ],
            pxrt.Width,
        )
    )
    def spec(
        self,
        transform_x,
        transform_N,
        transform_sign,
        transform_eps,
        transform_real,
        transform_ntrans,
        transform_nthreads,
        transform_modeord,
        request,
    ):
        ndi, width = request.param
        xp, dtype = ndi.module(), width.value
        with pxrt.Precision(width):
            op = pxo.NUFFT.type2(
                x=xp.array(transform_x, dtype=dtype),
                N=transform_N,
                isign=transform_sign,
                eps=transform_eps,
                real=transform_real,
                enable_warnings=False,
                n_trans=transform_ntrans,
                nthreads=transform_nthreads,
                modeord=transform_modeord,
            )
        return op, ndi, width

    @pytest.fixture
    def data_shape(
        self,
        transform_x,
        transform_N,
        transform_real,
    ):
        dim = np.prod(transform_N) * (1 if transform_real else 2)
        codim = 2 * len(transform_x)
        return (codim, dim)

    @pytest.fixture
    def data_apply(
        self,
        _transform_cArray,
        transform_real,
    ):
        M, N = _transform_cArray.shape

        rng = np.random.default_rng(3)
        w = rng.normal(size=(N,)) + 1j * rng.normal(size=(N,))
        if transform_real:
            w = w.real
        v = _transform_cArray @ w

        return dict(
            in_=dict(arr=w if transform_real else pxu.view_as_real(w)),
            out=pxu.view_as_real(v),
        )



import pyxu.abc as pxa
import pyxu.info.ptype as pxt
import pyxu.runtime as pxrt
import pyxu.util as pxu

__all__ = [
    "shift_loss",
    "KLDivergence",
]


def shift_loss(
    op: pxt.OpT,
    data: pxt.NDArray = None,
) -> pxt.OpT:
    r"""
    Shift a functional :math:`f(x)` to a loss functional.

    .. math::

       g(x; c) = f(x - c)

    Parameters
    ----------
    data: NDArray
        (M,) input data.

    Returns
    -------
    op: OpT
        (1, M) Loss functionial.  If `data` is omitted, then this function is a no-op.
    """
    if data is None:
        return op
    else:
        return op.argshift(-data)


class KLDivergence(pxa.ProxFunc):
    r"""
    Generalised Kullback-Leibler divergence :math:`D_{KL}(\mathbf{y}||\mathbf{x}):=\sum_{i=1}^N y_i\log(y_i/x_i) -y_i
    +x_i`.
    """

    def __init__(
        self,
        dim: pxt.Integer,
        data: pxt.NDArray,
    ):
        r"""
        Parameters
        ----------
        dim: Integer
        data: NDArray
            (M,) strictly positive input data.

        Examples
        --------
        .. code-block:: python3

           import numpy as np
           from pyxu.operator import KLDivergence

           y = np.arange(10)
           loss = KLDivergence(dim=y.size, data=y)
           x = 2 * np.arange(10)
           loss(x)
           # 13.80837687480246
           np.round(loss.prox(x, tau=1))
           # array([ 0.,  2.,  4.,  6.,  8., 10., 12., 14., 16., 18.])

        Notes
        -----
        In information theory, and in the case where :math:`\mathbf{y}` and :math:`\mathbf{x}`  sum to one  --and hence
        can be interpreted as discrete probability distributions, the KL-divergence can be interpreted as the relative
        entropy of :math:`\mathbf{y}` w.r.t. :math:`\mathbf{x}`, i.e. the amount of information lost when using
        :math:`\mathbf{x}` to approximate :math:`\mathbf{y}`. It is particularly useful in the context of count data
        with Poisson distribution. Indeed, the KL-divergence corresponds –up to an additive constant– to the likelihood
        of the data :math:`\mathbf{y}` where each component is independent with Poisson distribution and respective
        intensities given by the entries of :math:`\mathbf{x}`. See [FuncSphere]_ Section 5 of Chapter 7 for the
        computation of its proximal operator.
        """
        super().__init__(shape=(1, dim))

        xp = pxu.get_array_module(data)
        assert data.ndim == 1, "`data` must have 1 axis only."
        assert data.size == dim, "`data` must have the same size as the domain dimension."
        assert xp.all(data > 0), "KL Divergence only defined for positive-valued arguments."
        self.data = data.reshape(1, -1)

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        assert xp.all(arr > 0), "KL Divergence only defined for positive-valued arguments."
        sh = arr.shape[:-1]
        arr = arr.reshape(-1, self.data.size)
        out = xp.true_divide(self.data, arr)
        out = xp.log(out)
        out *= self.data
        out -= self.data
        out += arr
        return xp.sum(out, axis=-1).reshape(*sh, -1)

    @pxrt.enforce_precision(i=("arr", "tau"))
    def prox(self, arr: pxt.NDArray, tau: pxt.Real) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        sh = arr.shape[:-1]
        arr = arr.reshape(-1, self.data.size)
        out = xp.sqrt((arr - tau) ** 2 + 4 * tau * self.data)
        out += arr
        out -= tau
        out /= 2
        return out.reshape(*sh, -1)



import itertools

import pytest
import scipy.sparse.linalg as spsl

import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.operator.interop as pxio
import pyxu.runtime as pxrt
import pyxu_tests.operator.conftest as conftest


class FromSciOpMixin:
    # Fixtures ----------------------------------------------------------------
    @pytest.fixture
    def op_orig(self) -> pxt.OpT:
        # Override in inherited class with the operator used to create the sci_op.
        raise NotImplementedError

    @pytest.fixture(
        params=itertools.product(
            [
                pxd.NDArrayInfo.NUMPY,
                pxd.NDArrayInfo.CUPY,
                # DASK-based sci_ops unsupported
            ],
            pxrt.Width,
        )
    )
    def spec(self, op_orig, request) -> tuple[pxt.OpT, pxd.NDArrayInfo, pxrt.Width]:
        ndi, width = request.param
        self._skip_if_unsupported(ndi)

        if ndi == pxd.NDArrayInfo.CUPY:
            import cupyx.scipy.sparse.linalg as xpl
        else:  # NUMPY
            xpl = spsl

        A = op_orig.asarray(xp=ndi.module(), dtype=width.value)
        B = xpl.aslinearoperator(A)
        op = pxio.from_sciop(cls=self.base, sp_op=B)

        return op, ndi, width

    @pytest.fixture
    def data_shape(self, op_orig) -> pxt.OpShape:
        return op_orig.shape

    @pytest.fixture
    def data_apply(self, op_orig) -> conftest.DataLike:
        dim = op_orig.dim
        arr = self._random_array((dim,), seed=53)  # random seed for reproducibility
        out = op_orig.apply(arr)
        return dict(
            in_=dict(arr=arr),
            out=out,
        )

    @pytest.fixture
    def data_adjoint(self, op_orig) -> conftest.DataLike:
        codim = op_orig.codim
        arr = self._random_array((codim,), seed=54)  # random seed for reproducibility
        out = op_orig.adjoint(arr)
        return dict(
            in_=dict(arr=arr),
            out=out,
        )


class TestFromSciOpLinFunc(FromSciOpMixin, conftest.LinFuncT):
    @pytest.fixture
    def op_orig(self) -> pxt.OpT:
        import pyxu_tests.operator.examples.test_linfunc as tc

        return tc.ScaledSum(N=7)


class TestFromSciOpLinOp(FromSciOpMixin, conftest.LinOpT):
    @pytest.fixture
    def op_orig(self) -> pxt.OpT:
        import pyxu_tests.operator.examples.test_linop as tc

        return tc.Tile(N=3, M=4)


class TestFromSciOpSquareOp(FromSciOpMixin, conftest.SquareOpT):
    @pytest.fixture
    def op_orig(self) -> pxt.OpT:
        import pyxu_tests.operator.examples.test_squareop as tc

        return tc.CumSum(N=7)


class TestFromSciOpNormalOp(FromSciOpMixin, conftest.NormalOpT):
    @pytest.fixture
    def op_orig(self) -> pxt.OpT:
        import pyxu_tests.operator.examples.test_normalop as tc

        h = self._random_array((5,), seed=2)
        return tc.CircularConvolution(h=h)


class TestFromSciOpUnitOp(FromSciOpMixin, conftest.UnitOpT):
    @pytest.fixture
    def op_orig(self) -> pxt.OpT:
        import pyxu_tests.operator.examples.test_unitop as tc

        return tc.Permutation(N=7)


class TestFromSciOpSelfAdjointOp(FromSciOpMixin, conftest.SelfAdjointOpT):
    @pytest.fixture
    def op_orig(self) -> pxt.OpT:
        import pyxu_tests.operator.examples.test_selfadjointop as tc

        return tc.SelfAdjointConvolution(N=7)


class TestFromSciOpPosDefOp(FromSciOpMixin, conftest.PosDefOpT):
    @pytest.fixture
    def op_orig(self) -> pxt.OpT:
        import pyxu_tests.operator.examples.test_posdefop as tc

        return tc.PSDConvolution(N=7)


class TestFromSciOpProjOp(FromSciOpMixin, conftest.ProjOpT):
    @pytest.fixture
    def op_orig(self) -> pxt.OpT:
        import pyxu_tests.operator.examples.test_projop as tc

        return tc.Oblique(N=7, alpha=0.3)


class TestFromSciOpOrthProjOp(FromSciOpMixin, conftest.OrthProjOpT):
    @pytest.fixture
    def op_orig(self) -> pxt.OpT:
        import pyxu_tests.operator.examples.test_orthprojop as tc

        return tc.ScaleDown(N=7)



import collections.abc as cabc
import functools
import itertools
import operator

import numpy as np
import pytest

import pyxu.abc.operator as pxa
import pyxu.info.ptype as pxt
import pyxu.operator as pxo
import pyxu.opt.solver as pxsl
import pyxu_tests.opt.solver.conftest as conftest


def generate_funcs_K(descr, N_term) -> cabc.Sequence[tuple[pxt.OpT]]:
    # Take description of many functionals, i.e. output of funcs(), and return a stream of
    # length-N_term tuples, where each of the first N_term - 1 terms of each tuple is a functional created by composing
    # and summing a subset of `descr`, and the last term is a tuple (func, op) that is an element of `descr`.
    #
    # Examples
    # --------
    # generate_funcs([(a, b), (c, d), (e, f)], 2)
    # -> [ (c*d + e*f, (a, b)), (a*b + e*f, (c, d)), (a*b + c*d, (e, f)), ]
    # generate_funcs([(a, b), (c, d), (e, f)], 3)
    # -> [ (c*d, e*f, (a, b)), (a*b, e*f, (c, d)),
    #      (e*f, c*d, (a, b)), (a*b, c*d, (e, f)),
    #      (e*f, a*b, (c, d)), (c*d, a*b, (e, f)) ]

    assert 2 <= N_term <= len(descr)  # Must have at least 2 terms because h cannot be a sum

    def chain(x, y):
        comp = x * y
        comp.diff_lipschitz = comp.estimate_diff_lipschitz()
        return comp

    stream = []
    for d in itertools.permutations(descr, N_term - 1):
        to_sum = list(set(descr) - set(d))
        p = functools.reduce(operator.add, [chain(*dd) for dd in to_sum])
        p.diff_lipschitz = p.estimate_diff_lipschitz()
        stream.append([*d, p])  # The last subset is a func while the others are (func, op) tuples

    stream_K = []
    for part in stream:
        for n in range(N_term - 1):
            p = part.copy()
            del p[n]  # The n-th subset is used for the (h, K) pair
            p = [chain(*dd) for dd in p[:-1]]  # Compose all but the last subset (which is already a func)
            stream_K.append((*p, part[-1], part[n]))  # Append the last subset and the (h, K) tuple
    return stream_K


class MixinPDS(conftest.SolverT):
    disable_test = {
        # Several PDS tests fail with default parameters with certain combinations of objective functionals. They all
        # pass when removing the time threshold and adding a minimum number of iterations (to avoid solver stopping too
        # close to its initial point), but they are extremely slow.
        "test_value_fit",
    }

    dim = 5  # Dimension of input vector for tests

    @classmethod
    def generate_init_kwargs(cls, has_f: bool, has_g: bool, has_h: bool, has_K: bool) -> list[dict]:
        # Returns a stream of dictionaries for the init_kwargs fixture of the solver based on whether that solver has
        # arguments f, g, h and K. All possible combinations of the output of `funcs` are tested.

        funcs = conftest.funcs(cls.dim)
        stream1 = conftest.generate_funcs(funcs, N_term=1)
        stream2 = conftest.generate_funcs(funcs, N_term=2)

        kwargs_init = []
        if has_f:
            kwargs_init.extend([dict(f=f) for (f, *_) in stream1])
        if has_g:
            kwargs_init.extend([dict(g=g) for (g, *_) in stream1])
            if has_f:
                kwargs_init.extend([dict(f=f, g=g) for (f, g) in stream2])
        if has_h:
            kwargs_init.extend([dict(h=h) for (h, *_) in stream1])
            if has_f:
                kwargs_init.extend([dict(f=f, h=h) for (f, h) in stream2])
            if has_g:
                kwargs_init.extend([dict(g=g, h=h) for (g, h) in stream2])
                if has_f:
                    stream3 = conftest.generate_funcs(funcs, N_term=3)
                    kwargs_init.extend([dict(f=f, g=g, h=h) for (f, g, h) in stream3])

        if has_K:
            stream2_K = generate_funcs_K(funcs, N_term=2)
            if has_f:
                kwargs_init.extend([dict(f=f, h=h, K=K) for (f, (h, K)) in stream2_K])
            if has_g:
                kwargs_init.extend([dict(g=g, h=h, K=K) for (g, (h, K)) in stream2_K])
                if has_f:
                    stream3_K = generate_funcs_K(funcs, N_term=3)
                    kwargs_init.extend([dict(f=f, g=g, h=h, K=K) for (f, g, (h, K)) in stream3_K])
        return kwargs_init

    @staticmethod
    def fenchel_conjugate(quad_func: pxa.QuadraticFunc):
        # Fenchel conjugate of a quadratic function (up to the constant term)
        Q, c, _ = quad_func._quad_spec()
        Q_fench = Q.dagger()
        c_fench = -c * Q_fench
        return pxa.QuadraticFunc(shape=quad_func.shape, Q=Q_fench, c=c_fench)

    @pytest.fixture
    def spec(self, klass, init_kwargs, fit_kwargs) -> tuple[pxt.SolverC, dict, dict]:
        return klass, init_kwargs, fit_kwargs

    @pytest.fixture(params=[1, 2, 3])
    def tuning_strategy(self, request) -> int:
        return request.param

    @pytest.fixture(params=["CV", "PD3O"])
    def base(self, request) -> pxt.SolverC:
        bases = {"CV": pxsl.CV, "PD3O": pxsl.PD3O}
        return bases[request.param]

    @pytest.fixture(params=["1d", "nd"])
    def x0(self, request) -> dict:
        # Multiple initial points
        return {"1d": np.full((self.dim,), 3.0), "nd": np.full((2, self.dim), 15.0)}[request.param]

    @pytest.fixture
    def klass(self) -> pxt.SolverC:
        return NotImplementedError

    @pytest.fixture
    def init_kwargs(self) -> dict:
        return NotImplementedError

    @pytest.fixture
    def fit_kwargs(self, x0, tuning_strategy) -> dict:
        # Overriden only for ADMM
        return dict(
            x0=x0,
            tuning_strategy=tuning_strategy,
        )

    @pytest.fixture
    def cost_function(self, init_kwargs) -> dict[str, pxt.OpT]:
        kwargs = [init_kwargs.get(k, pxo.NullFunc(dim=self.dim)) for k in ("f", "g")]
        func = kwargs[0] + kwargs[1]
        out = dict()
        if init_kwargs.get("h") is not None:
            if func._name != "NullFunc":
                h = init_kwargs.get("h")
                K = init_kwargs.get("K", pxo.IdentityOp(dim=self.dim))
                if isinstance(h, pxa.QuadraticFunc) and isinstance(func, pxa.QuadraticFunc):
                    Q_h, _, _ = h._quad_spec()
                    Q_f, _, _ = func._quad_spec()
                    if (Q_h._name == "IdentityOp" or Q_h._name == "HomothetyOp") and (
                        Q_f._name == "IdentityOp" or Q_f._name == "HomothetyOp"
                    ):
                        # The dual cost is guaranteed to have finite values for QuadraticFuncs (it is infinite for
                        # LinFuncs). We only compute the dual cost for QuadraticFuncs with explicit inverses
                        out["z"] = self.fenchel_conjugate(h) + self.fenchel_conjugate(func) * (-K.T)
            func += h * K
        out["x"] = func
        return out


class TestPD3O(MixinPDS):
    @pytest.fixture
    def klass(self) -> pxt.SolverC:
        return pxsl.PD3O

    @pytest.fixture(params=MixinPDS.generate_init_kwargs(has_f=True, has_g=True, has_h=True, has_K=True))
    def init_kwargs(self, request) -> dict:
        return request.param


class TestCV(MixinPDS):
    @pytest.fixture
    def klass(self) -> pxt.SolverC:
        return pxsl.CV

    @pytest.fixture(params=MixinPDS.generate_init_kwargs(has_f=True, has_g=True, has_h=True, has_K=True))
    def init_kwargs(self, request) -> dict:
        return request.param


class TestCP(MixinPDS):
    @pytest.fixture
    def klass(self) -> pxt.SolverC:
        return pxsl.CP

    @pytest.fixture(params=MixinPDS.generate_init_kwargs(has_f=False, has_g=True, has_h=True, has_K=True))
    def init_kwargs(self, request, base) -> dict:
        kwargs = request.param
        kwargs.update({"base": base})
        return kwargs


class TestLV(MixinPDS):
    @pytest.fixture
    def klass(self) -> pxt.SolverC:
        return pxsl.LV

    @pytest.fixture(params=MixinPDS.generate_init_kwargs(has_f=True, has_g=False, has_h=True, has_K=True))
    def init_kwargs(self, request) -> dict:
        return request.param


class TestDY(MixinPDS):
    @pytest.fixture
    def klass(self) -> pxt.SolverC:
        return pxsl.DY

    @pytest.fixture(params=MixinPDS.generate_init_kwargs(has_f=True, has_g=True, has_h=True, has_K=False))
    def init_kwargs(self, request) -> dict:
        return request.param


class TestDR(MixinPDS):
    @pytest.fixture
    def klass(self) -> pxt.SolverC:
        return pxsl.DR

    @pytest.fixture(params=MixinPDS.generate_init_kwargs(has_f=False, has_g=True, has_h=True, has_K=False))
    def init_kwargs(self, request, base) -> dict:
        kwargs = request.param
        kwargs.update({"base": base})
        return kwargs


class TestADMM(MixinPDS):
    @staticmethod
    def nlcg_init_kwargs():
        # Returns init_kwargs to test the (otherwise untested) NLCG scenario
        funcs = conftest.funcs(MixinPDS.dim)

        def chain(x, y):
            comp = x * y
            comp.diff_lipschitz = comp.estimate_diff_lipschitz()
            return comp

        f = chain(*funcs.pop(2))  # f is a LinFunc (and not a QuadraticFunc) -> avoids CG scenario
        h = functools.reduce(operator.add, [chain(*ff) for ff in funcs])
        K = pxo.IdentityOp(dim=MixinPDS.dim)  # Hack to force NLCG scenario (if K = None, then classical ADMM)
        return dict(f=f, h=h, K=K)

    @pytest.fixture
    def klass(self) -> pxt.SolverC:
        return pxsl.ADMM

    @pytest.fixture(
        params=[*MixinPDS.generate_init_kwargs(has_f=True, has_g=False, has_h=True, has_K=True), nlcg_init_kwargs()]
    )
    def init_kwargs(self, request) -> dict:
        return request.param

    @pytest.fixture
    def spec(self, klass, init_kwargs, fit_kwargs) -> tuple[pxt.SolverC, dict, dict]:
        # Overriden from base class
        isNLCG = (init_kwargs.get("K") is not None) and (not isinstance(init_kwargs.get("f"), pxa.QuadraticFunc))
        if (fit_kwargs["x0"].squeeze().ndim > 1) and isNLCG:
            pytest.skip("NLCG scenario with multiple initial points not supported.")
        return klass, init_kwargs, fit_kwargs


class TestFB(MixinPDS):
    @pytest.fixture
    def klass(self) -> pxt.SolverC:
        return pxsl.FB

    @pytest.fixture(params=MixinPDS.generate_init_kwargs(has_f=True, has_g=True, has_h=False, has_K=False))
    def init_kwargs(self, request) -> dict:
        return request.param


class TestPP(MixinPDS):
    @pytest.fixture
    def klass(self) -> pxt.SolverC:
        return pxsl.PP

    @pytest.fixture(params=MixinPDS.generate_init_kwargs(has_f=False, has_g=True, has_h=False, has_K=False))
    def init_kwargs(self, request, base) -> dict:
        kwargs = request.param
        kwargs.update({"base": base})
        return kwargs



"""
Low-level functions used to define user-facing stencils.
"""
import collections.abc as cabc
import itertools
import string

import numpy as np

import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.runtime as pxrt


def _signature(params, returns) -> str:
    # Translate a signature of the form
    #     [in_1_spec, ..., in_N_spec] -> out_spec
    # to Numba's string representation.
    #
    # Parameters
    # ----------
    # params: list(spec)
    # returns: spec | None
    #
    # Returns
    # -------
    # sig: str
    #
    # Notes
    # -----
    # A parameter spec is characterized by the triplet
    #     (dtype[single/double], ndim[int], c_contiguous[bool])
    def fmt(spec) -> str:
        dtype, ndim, c_contiguous = spec

        _dtype_spec = {
            pxrt.Width.SINGLE: "float32",
            pxrt.Width.DOUBLE: "float64",
        }[pxrt.Width(dtype)]

        dim_spec = [":"] * ndim
        if c_contiguous and (ndim > 0):
            dim_spec[-1] = "::1"
        dim_spec = "[" + ",".join(dim_spec) + "]"

        _repr = _dtype_spec
        if ndim > 0:
            _repr += dim_spec
        return _repr

    sig = "".join(
        [
            "void" if (returns is None) else fmt(returns),
            "(",
            ", ".join(map(fmt, params)),
            ")",
        ]
    )
    return sig


class _Stencil:
    """
    Multi-dimensional JIT-compiled stencil. (Low-level function.)

    This low-level class creates a gu-vectorized stencil applicable on multiple inputs simultaneously.

    Create instances via factory method :py:meth:`~pyxu.operator._Stencil.init`.

    Example
    -------
    Correlate a stack of images `A` with a (3, 3) kernel such that:

    .. math::

       B[n, m] = A[n-1, m] + A[n, m-1] + A[n, m+1] + A[n+1, m]

    .. code-block:: python3

       import numpy as np
       from pyxu.operator.linop.stencil._stencil import _Stencil

       # create the stencil
       kernel = np.array([[0, 1, 0],
                          [1, 0, 1],
                          [0, 1, 0]], dtype=np.float64)
       center = (1, 1)
       stencil = _Stencil.init(kernel, center)

       # apply it to the data
       rng = np.random.default_rng()
       A = rng.normal(size=(2, 3, 4, 30, 30))  # 24 images of size (30, 30)
       B = np.zeros_like(A)
       stencil.apply(A, B)  # (2, 3, 4, 30, 30)
    """

    IndexSpec = cabc.Sequence[pxt.Integer]

    @staticmethod
    def init(
        kernel: pxt.NDArray,
        center: IndexSpec,
    ):
        """
        Parameters
        ----------
        kernel: NDArray
            (k_1, ..., k_D) kernel coefficients.

            Only float32/64 kernels are supported.
        center: ~pyxu.operator._Stencil.IndexSpec
            (D,) index of the kernel's center.

        Returns
        -------
        st: ~pyxu.operator.linop.stencil._stencil._Stencil
            Rank-D stencil.
        """
        dtype = kernel.dtype
        if dtype not in {_.value for _ in pxrt.Width}:
            raise ValueError(f"Unsupported kernel precision {dtype}.")

        center = np.array(center, dtype=int)
        assert center.size == kernel.ndim
        assert np.all(0 <= center) and np.all(center < kernel.shape)

        N = pxd.NDArrayInfo
        ndi = N.from_obj(kernel)
        if ndi == N.NUMPY:
            klass = _Stencil_NP
        elif ndi == N.CUPY:
            klass = _Stencil_CP
        else:
            raise NotImplementedError

        st = klass(kernel, center)
        return st

    def apply(
        self,
        arr: pxt.NDArray,
        out: pxt.NDArray,
        **kwargs,
    ) -> pxt.NDArray:
        r"""
        Evaluate stencil on multiple inputs.

        Parameters
        ----------
        arr: NDArray
            (..., N_1, ..., N_D) data to process.
        out: NDArray
            (..., N_1, ..., N_D) array to which outputs are written.
        kwargs: dict
            Extra kwargs to configure `f_jit()`, the Dispatcher instance created by Numba.

            Only relevant for GPU stencils, with values:

            * blockspergrid: int
            * threadsperblock: int

            Default values are chosen if unspecified.

        Returns
        -------
        out: NDArray
            (..., N_1, ..., N_D) outputs.

        Notes
        -----
        * `arr` and `out` must have the same type/dtype as the kernel used during instantiation.
        * Index regions in `out` where the stencil is not fully supported are set to 0.
        * :py:meth:`~pyxu.operator._Stencil.apply` may raise ``CUDA_ERROR_LAUNCH_OUT_OF_RESOURCES`` when the number of
          GPU registers required exceeds resource limits.  There are 2 solutions to this problem:

            (1) Pass the `max_registers` kwarg to f_jit()'s decorator; or
            (2) `Limit the number of threads per block <https://stackoverflow.com/a/68659008>`_.

          (1) must be set at compile time; it is thus left unbounded.
          (2) is accessible through .apply(\*\*kwargs).
        """
        assert arr.dtype == out.dtype == self._kernel.dtype
        assert arr.shape == out.shape
        # assert arr.flags.c_contiguous and out.flags.c_contiguous
        # [2022.12.25, Sepand]
        #   Preferable to explicitly enforce C-contiguity for better performance, but then
        #   LinOp.to_sciop() is partially broken.

        K_dim = len(self._kernel.shape)
        arg_shape = arr.shape[-K_dim:]

        stencil = self._configure_dispatcher(arr.size, **kwargs)
        stencil(
            arr.reshape(-1, *arg_shape),
            out.reshape(-1, *arg_shape),
        )
        return out

    def __init__(
        self,
        kernel: pxt.NDArray,
        center: pxt.NDArray,
    ):
        self._kernel = kernel
        self._center = center
        self._code = self._gen_code()

        exec(self._code, locals())  # compile stencil
        self._dispatch = eval("f_jit")  # keep track of JIT Dispatcher

    def _gen_code(self) -> str:
        # Generate code which creates `f_jit()` after execution.
        raise NotImplementedError

    def _configure_dispatcher(self, pb_size: int, **kwargs) -> cabc.Callable:
        # Configure `f_jit()`, the Numba Dispatcher instance.
        #
        # Parameters
        # ----------
        # pb_size: int
        #     Number of stencil evaluations.
        # **kwargs: dict
        #
        # Returns
        # -------
        # f: callable
        #     Configured Numba Dispatcher.
        raise NotImplementedError


class _Stencil_NP(_Stencil):
    def _gen_code(self) -> str:
        _template = string.Template(
            r"""
import numba

@numba.stencil(
    func_or_mode="constant",
    cval=0,
)
def f_stencil(a):
    # rank-D kernel [specified as rank-(D+1) to work across stacked dimension].
    return ${stencil_spec}

@numba.jit(
    "${signature}",
    nopython=True,
    nogil=True,
    forceobj=False,
    parallel=True,
    error_model="numpy",
    cache=False,  # not applicable to dynamically-defined functions (https://github.com/numba/numba/issues/3501)
    fastmath=True,
    boundscheck=False,
)
def f_jit(arr, out):
    # arr: (N_stack, N_1, ..., N_D)
    # out: (N_stack, N_1, ..., N_D)
    f_stencil(arr, out=out)
"""
        )
        code = _template.substitute(
            signature=self.__sig_spec(),
            stencil_spec=self.__stencil_spec(),
        )
        return code

    def _configure_dispatcher(self, pb_size: int, **kwargs) -> cabc.Callable:
        # Nothing to do for CPU targets.
        return self._dispatch

    def __sig_spec(self) -> str:
        sig_spec = (self._kernel.dtype, self._kernel.ndim + 1, False)
        signature = _signature([sig_spec, sig_spec], None)
        return signature

    def __stencil_spec(self) -> str:
        f_fmt = {  # coef float-formatter
            pxrt.Width.SINGLE: "1.8e",
            pxrt.Width.DOUBLE: "1.16e",
        }[pxrt.Width(self._kernel.dtype)]

        entry = []
        _range = list(map(range, self._kernel.shape))
        for idx in itertools.product(*_range):
            idx_c = [i - c for (i, c) in zip(idx, self._center)]
            idx_c = ",".join(map(str, idx_c))

            cst = self._kernel[idx]
            if np.isclose(cst, 0):
                # no useless look-ups at runtime
                e = None
            elif np.isclose(cst, 1):
                # no multiplication required
                e = f"a[0,{idx_c}]"
            else:
                # general case
                e = f"({cst:{f_fmt}} * a[0,{idx_c}])"

            if e is not None:
                entry.append(e)

        spec = " + ".join(entry)
        return spec


class _Stencil_CP(_Stencil):
    def _gen_code(self) -> str:
        _template = string.Template(
            r"""
import numba.cuda

@numba.cuda.jit(
    device=True,
    inline=True,
    fastmath=True,
    opt=True,
    cache=False,
)
def unravel_offset(
    offset,  # int
    shape,  # (D+1,)
):
    # Compile-time-equivalent of np.unravel_index(offset, shape, order='C').
    ${unravel_spec}
    return idx


@numba.cuda.jit(
    device=True,
    inline=True,
    fastmath=True,
    opt=True,
    cache=False,
)
def full_overlap(
    idx,  # (D+1,)
    dimensions,  # (D+1,)
):
    # Computes intersection of:
    # * idx[0] < dimensions[0]
    # * 0 <= idx[1:] - kernel_center
    # * idx[1:] - kernel_center <= dimensions[1:] - kernel_width
    if not (idx[0] < dimensions[0]):
        # went beyond stack-dim limit
        return False

    kernel_width = ${kernel_width}
    kernel_center = ${kernel_center}
    for i, w, c, n in zip(
        idx[1:],
        kernel_width,
        kernel_center,
        dimensions[1:],
    ):
        if not (0 <= i - c <= n - w):
            # kernel partially overlaps ROI around `idx`
            return False

    # kernel fully overlaps ROI around `idx`
    return True


@numba.cuda.jit(
    func_or_sig="${signature}",
    device=False,
    inline=False,
    fastmath=True,
    opt=True,
    cache=False,  # not applicable to dynamically-defined functions (https://github.com/numba/numba/issues/3501)
    # max_registers=None,  # see .apply() notes
)
def f_jit(arr, out):
    # arr: (N_stack, N_1, ..., N_D)
    # out: (N_stack, N_1, ..., N_D)
    offset = numba.cuda.grid(1)
    if offset < arr.size:
        idx = unravel_offset(offset, arr.shape)
        if full_overlap(idx, arr.shape):
            out[idx] = ${stencil_spec}
        else:
            out[idx] = 0
"""
        )
        code = _template.substitute(
            kernel_center=str(tuple(self._center)),
            kernel_width=str(self._kernel.shape),
            signature=self.__sig_spec(),
            stencil_spec=self.__stencil_spec(),
            unravel_spec=self.__unravel_spec(),
        )
        return code

    def _configure_dispatcher(self, pb_size: int, **kwargs) -> cabc.Callable:
        # Set (`threadsperblock`, `blockspergrid`)
        assert set(kwargs.keys()) <= {
            "threadsperblock",
            "blockspergrid",
        }

        attr = self._kernel.device.attributes
        tpb = kwargs.get("threadsperblock", attr["MaxThreadsPerBlock"])
        bpg = kwargs.get("blockspergrid", (pb_size // tpb) + 1)
        return self._dispatch[bpg, tpb]

    def __sig_spec(self) -> str:
        sig_spec = (self._kernel.dtype, self._kernel.ndim + 1, False)
        signature = _signature([sig_spec, sig_spec], None)
        return signature

    def __stencil_spec(self) -> str:
        f_fmt = {  # coef float-formatter
            pxrt.Width.SINGLE: "1.8e",
            pxrt.Width.DOUBLE: "1.16e",
        }[pxrt.Width(self._kernel.dtype)]

        entry = []
        _range = list(map(range, self._kernel.shape))
        for idx in itertools.product(*_range):
            # create string of form "idx[1]+i1,...,idx[K]+iK"
            idx_c = [i - c for (i, c) in zip(idx, self._center)]
            idx_c = [f"idx[{i1}]{i2:+d}" for (i1, i2) in enumerate(idx_c, start=1)]
            idx_c = ",".join(idx_c)

            cst = self._kernel[idx]
            if np.isclose(cst, 0):
                # no useless look-ups at runtime
                e = None
            elif np.isclose(cst, 1):
                # no multiplication required
                e = f"arr[idx[0],{idx_c}]"
            else:
                # general case
                e = f"({cst:{f_fmt}} * arr[idx[0],{idx_c}])"

            if e is not None:
                entry.append(e)

        spec = " + ".join(entry)
        return spec

    def __unravel_spec(self) -> str:
        N = self._kernel.ndim + 1  # 1 stack-dim
        entry = []

        # left = offset
        e = "left = offset"
        entry.append(e)

        # blk = prod(shape)
        e = "blk = " + " * ".join([f"shape[{n}]" for n in range(N)])
        entry.append(e)

        for n in range(N):
            # blk //= shape[n]
            e = f"blk //= shape[{n}]"
            entry.append(e)

            # i{n} = left // blk
            e = f"i{n} = left // blk"
            entry.append(e)

            # left -= i{n} * blk
            e = f"left -= i{n} * blk"
            entry.append(e)

        # idx = (i0, ..., i{N})
        e = "idx = (" + ", ".join([f"i{n}" for n in range(N)]) + ")"
        entry.append(e)

        # indent each entry by 4, then concatenate
        for i in range(1, len(entry)):  # 1st line skipped
            entry[i] = "    " + entry[i]
        spec = "\n".join(entry)
        return spec



# How ArgShiftRule tests work:
#
# * ArgShiftRuleMixin auto-defines all arithmetic method (input,output) pairs.
#   [Caveat: we assume all tested examples are defined on \bR.] (This is not a problem in practice.)
#   [Caveat: we assume the base operators (op_orig) are correctly implemented.] (True if choosing test operators from examples/.)
#
# * To test an arg-shifted-operator, inherit from ArgShiftRuleMixin and the suitable MapT subclass
#   which the arg-shifted operator should abide by.
#
# Important: argshifted-operators are not module/precision-agnostic!

import collections.abc as cabc
import itertools
import typing as typ

import numpy as np
import pytest

import pyxu.abc as pxa
import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.runtime as pxrt
import pyxu_tests.operator.conftest as conftest
import pyxu_tests.operator.examples.test_diffmap as tc_diffmap
import pyxu_tests.operator.examples.test_linfunc as tc_linfunc
import pyxu_tests.operator.examples.test_linop as tc_linop
import pyxu_tests.operator.examples.test_normalop as tc_normalop
import pyxu_tests.operator.examples.test_orthprojop as tc_orthprojop
import pyxu_tests.operator.examples.test_posdefop as tc_posdefop
import pyxu_tests.operator.examples.test_projop as tc_projop
import pyxu_tests.operator.examples.test_proxdifffunc as tc_proxdifffunc
import pyxu_tests.operator.examples.test_selfadjointop as tc_selfadjointop
import pyxu_tests.operator.examples.test_squareop as tc_squareop
import pyxu_tests.operator.examples.test_unitop as tc_unitop

rng = np.random.default_rng(seed=50)


class ArgShiftRuleMixin:
    # Fixtures ----------------------------------------------------------------
    @pytest.fixture
    def op_orig(self) -> pxt.OpT:
        # Override in inherited class with the operator to be arg-shifted.
        raise NotImplementedError

    @pytest.fixture(params=[True, False])
    def op_shift(self, op_orig, request) -> typ.Union[pxt.Real, pxt.NDArray]:
        # Arg-shift values applied to op_orig()
        dim = self._sanitize(op_orig.dim, 7)
        cst = self._random_array((dim,), seed=20)  # random seed for reproducibility
        if request.param:  # scalar output
            cst = float(cst.sum())
        return cst

    @pytest.fixture(
        params=itertools.product(
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def spec(self, op_orig, op_shift, request) -> tuple[pxt.OpT, pxd.NDArrayInfo, pxrt.Width]:
        ndi, width = request.param
        self._skip_if_unsupported(ndi)
        if isinstance(op_shift, pxt.Real):
            shift = op_shift
        else:
            xp = ndi.module()
            shift = xp.array(op_shift, dtype=width.value)
        op = op_orig.argshift(shift)
        return op, ndi, width

    @pytest.fixture
    def data_shape(self, op_orig) -> pxt.OpShape:
        return op_orig.shape

    @pytest.fixture
    def data_apply(self, op_orig, op_shift) -> conftest.DataLike:
        dim = self._sanitize(op_orig.dim, 7)
        arr = self._random_array((dim,), seed=20)  # random seed for reproducibility
        out = op_orig.apply(arr + op_shift)
        return dict(
            in_=dict(arr=arr),
            out=out,
        )

    @pytest.fixture
    def data_grad(self, op_orig, op_shift) -> conftest.DataLike:
        dim = self._sanitize(op_orig.dim, 7)
        arr = self._random_array((dim,), seed=20)  # random seed for reproducibility
        out = op_orig.grad(arr + op_shift)
        return dict(
            in_=dict(arr=arr),
            out=out,
        )

    @pytest.fixture
    def data_prox(self, op_orig, op_shift) -> conftest.DataLike:
        dim = self._sanitize(op_orig.dim, 7)
        arr = self._random_array((dim,), seed=20)  # random seed for reproducibility
        tau = np.abs(self._random_array((1,), seed=21))[0]  # random seed for reproducibility
        out = op_orig.prox(arr + op_shift, tau) - op_shift
        return dict(
            in_=dict(
                arr=arr,
                tau=tau,
            ),
            out=out,
        )

    @pytest.fixture
    def data_math_lipschitz(self, op) -> cabc.Collection[np.ndarray]:
        N_test, dim = 5, self._sanitize(op.dim, 7)
        return self._random_array((N_test, dim))

    @pytest.fixture
    def data_math_diff_lipschitz(self, op) -> cabc.Collection[np.ndarray]:
        N_test, dim = 5, self._sanitize(op.dim, 7)
        return self._random_array((N_test, dim))

    # Tests -------------------------------------------------------------------
    @pytest.mark.skip("undefined for argshift.")
    def test_interface_asloss(self, op, xp, width):
        pass


# Test classes (Maps) ---------------------------------------------------------
class TestArgShiftRuleMap(ArgShiftRuleMixin, conftest.MapT):
    @pytest.fixture
    def op_orig(self):
        import pyxu_tests.operator.examples.test_map as tc

        return tc.ReLU(M=6)


class TestArgShiftRuleDiffMap(ArgShiftRuleMixin, conftest.DiffMapT):
    @pytest.fixture(
        params=[
            tc_diffmap.Sin(M=6),
            tc_linop.Tile(N=3, M=4),
            tc_squareop.CumSum(N=7),
            tc_normalop.CircularConvolution(h=rng.normal(size=(5,))),
            tc_unitop.Permutation(N=7),
            tc_selfadjointop.SelfAdjointConvolution(N=7),
            tc_posdefop.PSDConvolution(N=7),
            tc_projop.Oblique(N=6, alpha=np.pi / 4),
            tc_orthprojop.ScaleDown(N=7),
        ]
    )
    def op_orig(self, request):
        return request.param


# Test classes (Funcs) --------------------------------------------------------
class TestArgShiftRuleFunc(ArgShiftRuleMixin, conftest.FuncT):
    @pytest.fixture(params=[7])
    def op_orig(self, request):
        import pyxu_tests.operator.examples.test_func as tc

        return tc.Median(dim=request.param)


class TestArgShiftRuleDiffFunc(ArgShiftRuleMixin, conftest.DiffFuncT):
    @pytest.fixture(params=[7])
    def op_orig(self, request):
        import pyxu_tests.operator.examples.test_difffunc as tc

        return tc.SquaredL2Norm(M=request.param)


class TestArgShiftRuleProxFunc(ArgShiftRuleMixin, conftest.ProxFuncT):
    @pytest.fixture(params=[7])
    def op_orig(self, request):
        import pyxu_tests.operator.examples.test_proxfunc as tc

        return tc.L1Norm(M=request.param)


class TestArgShiftRuleQuadraticFunc(ArgShiftRuleMixin, conftest.QuadraticFuncT):
    @pytest.fixture(params=[0, 1])
    def op_orig(self, request):
        from pyxu_tests.operator.examples.test_linfunc import ScaledSum
        from pyxu_tests.operator.examples.test_posdefop import PSDConvolution

        N = 7
        op = {
            0: pxa.QuadraticFunc(shape=(1, N)),
            1: pxa.QuadraticFunc(
                shape=(1, N),
                Q=PSDConvolution(N=N),
                c=ScaledSum(N=N),
                t=1,
            ),
        }[request.param]
        return op


class TestArgShiftRuleProxDiffFunc(ArgShiftRuleMixin, conftest.ProxDiffFuncT):
    @pytest.fixture(
        params=[
            tc_proxdifffunc.SquaredL2Norm(M=7),
            tc_linfunc.ScaledSum(N=7),
        ]
    )
    def op_orig(self, request):
        return request.param



import itertools

import numpy as np
import pytest

import pyxu.info.ptype as pxt
import pyxu.opt.solver as pxsl
import pyxu.runtime as pxrt
import pyxu.util as pxu
import pyxu_tests.conftest as ct
import pyxu_tests.opt.solver.conftest as conftest


class TestCG(conftest.SolverT):
    @staticmethod
    def spec_data(N: int) -> list[tuple[pxt.SolverC, dict, dict]]:
        from pyxu_tests.operator.examples.test_posdefop import PSDConvolution

        klass = [
            pxsl.CG,
        ]
        kwargs_init = [
            dict(A=PSDConvolution(N=N)),
        ]

        kwargs_fit = []
        param_sweep = dict(
            b=[
                np.ones((N,)),
                np.full((2, N), -2),  # multiple problems in parallel
            ],
            x0=[
                None,  # let algorithm choose
                np.full((N,), 30),
            ],
            restart_rate=[None, N, 2 * N],
        )
        for config in itertools.product(*param_sweep.values()):
            d = dict(zip(param_sweep.keys(), config))
            kwargs_fit.append(d)

        data = itertools.product(klass, kwargs_init, kwargs_fit)
        return list(data)

    @pytest.fixture(params=spec_data(N=7))
    def spec(self, request) -> tuple[pxt.SolverC, dict, dict]:
        klass, kwargs_init, kwargs_fit = request.param
        return klass, kwargs_init, kwargs_fit

    # Note the different signature compared to SolverT.cost_function():
    # we need access to `b` from `kwargs_fit` to compute the cost function.
    # Moreover, this cost_function is NOT backend-agnostic.
    @pytest.fixture
    def cost_function(self, kwargs_init, kwargs_fit) -> dict[str, pxt.OpT]:
        import pyxu.abc as pxa
        import pyxu.operator as pxo

        # The value of `b` determines the cost function.
        # Moreover several `b` may be provided.
        # A unique cost function is obtained by flattening `b` and block-expanding the quadratic
        # cost accordingly.

        A = kwargs_init["A"]
        b = kwargs_fit["b"]
        N_b = int(np.prod(b.shape[:-1]))
        func = pxa.QuadraticFunc(
            shape=(1, N_b * A.dim),
            Q=pxo.block_diag((A,) * N_b),
            c=pxa.LinFunc.from_array(-b.reshape(-1), enable_warnings=False),
        )
        return dict(x=func)

    # Note the different signature compared to SolverT.test_value_fit().
    def test_value_fit(
        self,
        solver,
        _kwargs_fit_xp,
        width,
        cost_function,
        ground_truth,
    ):
        self._skip_if_disabled()

        # Due to different b/x0 domain combinations, it is difficult to use SolverT.test_value_fit().
        # To test convergence, we just check that the (strongly-convex) cost function is minimized.
        with pxrt.Precision(width):
            solver.fit(**_kwargs_fit_xp.copy())
        data, _ = solver.stats()

        cost_value = dict()
        for var, c_func in cost_function.items():
            x_gt = ground_truth[var]
            c_gt = c_func.apply(pxu.to_NUMPY(x_gt))
            # cost_function defined for NUMPY inputs only. [CG-peculiarity.]

            x_opt = data[var].reshape(-1)
            c_opt = c_func.apply(pxu.to_NUMPY(x_opt))
            # cost_function defined for NUMPY inputs only. [CG-peculiarity.]

            cost_value[var] = ct.allclose(c_gt, c_opt, as_dtype=width.value)
        assert all(cost_value.values())
